# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_nbs.ipynb.

# %% auto 0
__all__ = ['NBDEV_DEFAULT_EXP', 'AUTORELOAD', 'NBDEV_SHOWDOC', 'NBDEV_EXPORT', 'find_notebook', 'NotebookLoader',
           'NotebookFinder', 'NotebookViewer', 'NotebookPrefixer', 'NotebookAggregator', 'NotebookLogger']

# %% ../nbs/00_nbs.ipynb 4
import logging, inspect

# %% ../nbs/00_nbs.ipynb 7
import io, os, sys, types
from IPython import get_ipython
from nbformat import read
from IPython.core.interactiveshell import InteractiveShell

# %% ../nbs/00_nbs.ipynb 8
from pygments import highlight
from pygments.lexers import PythonLexer
from pygments.formatters import HtmlFormatter

from IPython.display import display, HTML

# %% ../nbs/00_nbs.ipynb 9
from typing import Optional
from dataclasses import dataclass, field

# %% ../nbs/00_nbs.ipynb 11
def find_notebook(fullname:str, path:Optional[str]=None) -> str:
    """find a notebook, given its fully qualified name and an optional path

    This turns "foo.bar" into "foo/bar.ipynb"
    and tries turning "Foo_Bar" into "Foo Bar" if Foo_Bar
    does not exist.
    """
    name = fullname.rsplit('.', 1)[-1]
    if not path:
        path = ['']

    for d in path:        
        nb_path = os.path.join(d, name + ".ipynb")
        if os.path.isfile(nb_path):
            return nb_path
        
        # let import Notebook_Name find "Notebook Name.ipynb"
        nb_path = nb_path.replace("_", " ")
        if os.path.isfile(nb_path):            
            return nb_path

# %% ../nbs/00_nbs.ipynb 13
class NotebookLoader(object):
    """Module Loader for Jupyter Notebooks"""
    def __init__(self, path:Optional[str]=None):        
        self.shell = InteractiveShell.instance()
        self.path = path

    def load_module(self, fullname:str):
        """import a notebook as a module"""
        path = find_notebook(fullname, self.path)

        print("importing Jupyter notebook from %s" % path)

        # load the notebook object
        with io.open(path, 'r', encoding='utf-8') as f:
            nb = read(f, 4)


        # create the module and add it to sys.modules
        # if name in sys.modules:
        #    return sys.modules[name]
        mod = types.ModuleType(fullname)
        mod.__file__ = path
        mod.__loader__ = self
        mod.__dict__['get_ipython'] = get_ipython
        sys.modules[fullname] = mod

        # extra work to ensure that magics that would affect the user_ns
        # actually affect the notebook module's ns
        save_user_ns = self.shell.user_ns
        self.shell.user_ns = mod.__dict__

        try:
          for cell in nb.cells:
            if cell.cell_type == 'code':
                # transform the input to executable Python
                code = self.shell.input_transformer_manager.transform_cell(cell.source)
                # run the code in themodule
                exec(code, mod.__dict__)

        finally:
            self.shell.user_ns = save_user_ns
            
        return mod

# %% ../nbs/00_nbs.ipynb 15
class NotebookFinder(object):
    """Module finder that locates Jupyter Notebooks"""
    def __init__(self):
        self.loaders = {}

    def find_module(self, fullname:str, path:Optional[str]=None) -> Optional[NotebookLoader]:
        nb_path = find_notebook(fullname, path)
        if not nb_path:
            return

        key = path
        if path:
            # lists aren't hashable
            key = os.path.sep.join(path)

        if key not in self.loaders:
            self.loaders[key] = NotebookLoader(path)
        return self.loaders[key]

# %% ../nbs/00_nbs.ipynb 19
@dataclass
class NotebookViewer(object):
    path: str

    lexer: PythonLexer = PythonLexer()
    formatter: HtmlFormatter = HtmlFormatter()

    def __init__(self, path:Optional[str]=None):              
        self.path = path

    def __post_init__(self):
        # self.formatter = HtmlFormatter()

        # self.lexer = PythonLexer()
        # publish the CSS for pygments highlighting
        style_defs = self.formatter.get_style_defs()
        self.style = """<style type='text/css'>%s</style>""" % style_defs
        display(HTML(self.style))

    def get_cell(self, idx:int, fname:Optional[str]=None):
        path = fname if fname else self.path        
        with io.open(path, 'r', encoding='utf-8') as f:
            nb = read(f, 4)

        cell = nb.cells[idx]
        return cell

    def prepare_cell(self, idx:int, fname:Optional[str]=None):
        cell = self.get_cell(idx, fname)

        html = []
        
        html.append("<h4>%s cell</h4>" % cell.cell_type)
        if cell.cell_type == 'code':
            html.append(highlight(cell.source, self.lexer, self.formatter))            
        else:
            html.append("<pre>%s</pre>" % cell.source)
        return html, cell.cell_type, cell.source
    
    def show_cell(self, idx:int, fname:Optional[str]=None):
        html, cell_type, cell_source = self.prepare_cell(idx, fname)
        return display(HTML('\n'.join(html)))
    
    def show_notebook(self, fname:Optional[str]=None):
        path = fname if fname else self.path
        """display a short summary of the cells of a notebook"""
        with io.open(path, 'r', encoding='utf-8') as f:
            nb = read(f, 4)

        html = []
        for cell in nb.cells:
            html.append("<h4>%s cell</h4>" % cell.cell_type)
            if cell.cell_type == 'code':
                html.append(highlight(cell.source, self.lexer, self.formatter))
            else:
                html.append("<pre>%s</pre>" % cell.source)
        return display(HTML('\n'.join(html)))

# %% ../nbs/00_nbs.ipynb 23
import os, re, nbformat as nbf
from nbformat import NotebookNode
from typing import List, Optional, Union
from tqdm.auto import tqdm
import warnings

# %% ../nbs/00_nbs.ipynb 24
@dataclass
class NotebookPrefixer:
    path: Optional[str] = None
    name: Optional[str] = None
    maximal: Optional[bool] = False

    def __post_init__(self):
        self.path = self.path if self.path else os.getcwd()


    def int_to_prefix_int(self, i: int) -> str:
        if i < 10:
            return f'0{i}'
        else:
            return str(i)
        
    def prefix_int_to_int(self, prefix: str) -> int:
        if prefix == '00':
            return 0        
        return int(prefix.lstrip('0'))
        
    def prefix_from_filename(self, name:str):
        return self.prefix_int_to_int(name.split('_')[0])
    
    def filename_without_prefix(self, name:str):
        if '_' not in name:
            return name
        return '_'.join(name.split('_')[1:])

    def next_nb_prefix(
        self, path: Optional[str] = None, 
        name: Optional[str] = None, maximal: bool = False
    ):
        """
        Calculate the next missing value of the notebook prefixes in a directory.

        Parameters
        ----------
        path : str
            The path to the directory containing the notebooks.
        name : Optional[str], optional
            The name of the notebook to check, by default None
        maximal : bool, optional
            If True, return the maximum prefix + 1, otherwise return the smallest missing prefix, by default False

        Returns
        -------
        int
            The next missing value of the notebook prefixes.
        """
        # Get the list of notebooks in the directory
        path = path if path else self.path
        name = name if name else self.name
        maximal = maximal if maximal is not None else self.maximal

        notebooks = [file for file in os.listdir(path) if file.endswith('.ipynb')]

        # Filter out non-convention notebooks
        notebooks = [file for file in notebooks if re.match(r'\d{2,}_.*\.ipynb', file)]

        # Extract the prefixes
        prefixes = sorted([self.prefix_from_filename(file) for file in notebooks])
        # Check if a name is provided
        if name is not None:
            # Check if there is a notebook that matches the name
            for file in notebooks:
                cur_name = self.filename_without_prefix(file).replace('.ipynb', '')
                basename = self.filename_without_prefix(name).replace('.ipynb', '')
                if basename == cur_name:
                    prefix = self.prefix_from_filename(file)
                    prefix = self.int_to_prefix_int(prefix)
                    return prefix

        if maximal:
            # Return the maximum prefix + 1
            return self.int_to_prefix_int(max(prefixes) + 1)
        
        else:
            # Return the smallest missing prefix
            for i, prefix in enumerate(prefixes[:-1]):
                nextix = prefixes[i + 1]
                if nextix - prefix > 1:
                    return self.int_to_prefix_int(prefix + 1)
                

# %% ../nbs/00_nbs.ipynb 26
NBDEV_DEFAULT_EXP = '#| default_exp'
AUTORELOAD = '%load_ext autoreload\n%autoreload 2'
NBDEV_SHOWDOC = '#| hide\nfrom nbdev.showdoc import *'
NBDEV_EXPORT = '#| hide\nimport nbdev; nbdev.nbdev_export()'

# %% ../nbs/00_nbs.ipynb 27
@dataclass
class NotebookAggregator:
    """
    A class to aggregate Jupyter notebooks.

    Parameters
    ----------
    path : str
        The path to the directory containing the notebooks.
    module : Optional[str], optional
        The name of the module, by default None
    output : Optional[str], optional
        The path to the output notebook, by default None
    ignore : List[str], optional
        A list of notebooks to ignore, by default []
    """
    path: str

    module: Optional[str] = None
    output: Optional[str] = None
    ignore: List[str] = field(default_factory=list)

    prefix: Optional[Union[bool, str]] = True
    prefix_dir: Optional[str] = None
    
    @property
    def basename(self):
        return os.path.basename(self.path)

    @property
    def dirname(self):
        return os.path.dirname(self.path)
    
    @property
    def notebooks(self):
        notebooks = [
            file for file in os.listdir(self.path) 
            if file.endswith('.ipynb') and file not in self.ignore
        ]
        notebooks = sorted(notebooks)
        return notebooks

    def __post_init__(self):
        if not self.module:
            self.module = self.basename

        if not self.prefix_dir:
            self.prefix_dir = self.dirname

        if not self.output:
            prefix = ''
            if self.prefix == True:
                prefix = f'{self.get_prefix()}_'

            elif isinstance(self.prefix, str):
                prefix = f'{self.prefix}_'
            
            self.output = os.path.join(self.dirname, f'{prefix}{self.module}.ipynb')

    def nb_as_href(self, name: str):
        return f'[{name}]({os.path.join(self.path, name)})'

    def format_header(self):
        res = f'# {self.module}\n'
        res += '> This notebook was generated from the following notebooks:\n\n'
        for name in self.notebooks:
            res += f'- {self.nb_as_href(name)}\n'
        return res
    
    def add_current_notebook_section(self, nb: NotebookNode, name: str):
        head = name.lstrip('_').rstrip('.ipynb').replace('_', ' ')
        if head.split()[0].isnumeric():
            head = ' '.join(head.split()[1:])

        nb['cells'].append(nbf.v4.new_markdown_cell(
            f'## {head.capitalize()}\n> This notebook was generated from {self.nb_as_href(name)})'
        ))
        return nb
    
    def write_nbdev_head(self, nb: NotebookNode):
        # Add the first four cells
        nb['cells'].append(nbf.v4.new_markdown_cell(self.format_header()))
        nb['cells'].append(nbf.v4.new_code_cell(f'{NBDEV_DEFAULT_EXP} {self.module}'))
        nb['cells'].append(nbf.v4.new_code_cell(AUTORELOAD))
        nb['cells'].append(nbf.v4.new_code_cell(NBDEV_SHOWDOC))
        return nb
    
    def write_nbdev_tail(self, nb: NotebookNode):  
        # Add the final cell      
        nb['cells'].append(nbf.v4.new_code_cell(NBDEV_EXPORT))
        return nb
    
    def get_prefix(self):
        prefixer = NotebookPrefixer(path=self.prefix_dir, name=self.module)
        prefix = prefixer.next_nb_prefix()
        return prefix

    def aggregate(self):
        new_nb = nbf.v4.new_notebook()
        new_nb['cells'] = []

        new_nb = self.write_nbdev_head(new_nb)

        # Iterate over the notebooks
        for nb_name in tqdm(self.notebooks, desc='Aggregating notebooks'):

            # Read the notebook
            with open(os.path.join(self.path, nb_name)) as f:
                nb = nbf.read(f, as_version=4)

            # Find the start and end cells
            start_cell = next(i for i, cell in enumerate(nb.cells) if cell.source.strip() == NBDEV_SHOWDOC)
            end_cell = next(i for i, cell in enumerate(nb.cells) if cell.source.strip() == NBDEV_EXPORT)

            self.add_current_notebook_section(new_nb, nb_name)

            # Add the cells to the new notebook
            new_nb.cells.extend(nb.cells[start_cell+1:end_cell])

        new_nb = self.write_nbdev_tail(new_nb)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            # Normalize the notebook
            changes, new_nb = nbf.validator.normalize(new_nb)

        # Write the new notebook
        with open(self.output, 'w') as f:
            nbf.write(new_nb, f)

# %% ../nbs/00_nbs.ipynb 33
import logging, inspect

# %% ../nbs/00_nbs.ipynb 34
@dataclass
class NotebookLogger(logging.Logger):
    name: Optional[str]  = 'NotebookLogger'
    level: Optional[int] = logging.INFO
    
    format: Optional[str] = '%(asctime)s - %(levelname)s - %(message)s'
    
    def __post_init__(self, *args, **kwargs):        
        if isinstance(self.level, str):
            self.level = logging.getLevelNamesMapping()[self.level.upper()]

        defaults = dict(name=self.name, level=self.level)
        kwargs.update(defaults)
        params = inspect.signature(logging.Logger.__init__).parameters
        kwargs = {k:v for k,v in kwargs.items() if k in params}        
        super().__init__(**kwargs)

        self.formatter = logging.Formatter(self.format)

        self._stdout = sys.stdout
        self._stderr = sys.stderr

        # Add the stream handler to the Jupyter logger
        self._stream_handler = logging.StreamHandler(self._stdout)
        self._stream_handler.setFormatter(self.formatter)
        self.addHandler(self._stream_handler)
        
        self._stderr_handler = logging.StreamHandler(self._stderr)
        self._stderr_handler.setLevel(logging.ERROR)
        self._stderr_handler.setFormatter(self.formatter)
        self.addHandler(self._stderr_handler)

    def __del__(self):
        self.removeHandler(self._stream_handler)
        self.removeHandler(self._stderr_handler)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/97_rich.ipynb.

# %% auto 0
__all__ = ['console', 'swrap', 'cprint']

# %% ../nbs/97_rich.ipynb 3
from rich.table import Table
from rich.console import Console

# %% ../nbs/97_rich.ipynb 4
console = Console()

# %% ../nbs/97_rich.ipynb 5
def swrap(style:str, text:str) -> str:
    return f'[{style}]{text}[/{style}]'

def cprint(style:str, text:str) -> None:
    console.print(swrap(style, text))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_types.ipynb.

# %% auto 0
__all__ = ['Series', 'ndarray', 'NPArray', 'DataFrame', 'DataFrames', 'Numeric', 'Bool', 'Boolish', 'IndexLike', 'SeriesLike',
           'IterLike', 'GroupKey', 'PathType', 'PathLike', 'SplitTuple2', 'SplitTuple3', 'SplitAsList',
           'DatasetSplitSpec', 'SplitDataset2Idxs', 'SplitDataset3Idxs', 'SplitDataFrames2', 'SplitDataFrames3',
           'SplitDatasets', 'ColorMap', 'TreeEntryFunc', 'RichTree', 'RichText', 'RichConsole', 'RichProgress',
           'TrainValidSplit', 'TrainTestSplit', 'TrainValidTestSplit', 'DirectoryTreeStrings']

# %% ../nbs/03_types.ipynb 5
from typing import (Tuple, Union, TypeAlias, List, Iterable, Literal, Any)
from numbers import Number

# %% ../nbs/03_types.ipynb 6
import numpy as np, pandas as pd
import pathlib

# %% ../nbs/03_types.ipynb 9
try:
    import torch
    Tensor: TypeAlias = torch.Tensor
    Device: TypeAlias = torch.device
except ImportError:
    # Define a placeholder type if torch is not installed
    Tensor: TypeAlias = Union[Iterable[Number], Any]
    Device: TypeAlias = Union[str, None, Any]
    pass

# %% ../nbs/03_types.ipynb 11
try:
    import anndata as ad    
    AnnData: TypeAlias = ad.AnnData
    AnnDatas: TypeAlias = List[AnnData]
except ImportError:
    # Define a placeholder type if torch is not installed
    AnnData: TypeAlias = Any
    AnnDatas: TypeAlias = List[AnnData]
    pass

# %% ../nbs/03_types.ipynb 12
Series: TypeAlias = pd.Series

ndarray: TypeAlias = np.ndarray

NPArray: TypeAlias = np.ndarray

DataFrame: TypeAlias = pd.DataFrame

DataFrames: TypeAlias = List[DataFrame]

# %% ../nbs/03_types.ipynb 13
Numeric: TypeAlias = Union[int, float, complex, Number]

Bool: TypeAlias = Union[bool, np.bool_]

Boolish: TypeAlias = Union[Bool, Literal[0], Literal[1]]

IndexLike: TypeAlias = Union[List[int], List[Boolish], pd.Index]

SeriesLike: TypeAlias = Union[list, Series, ndarray, Tensor, IndexLike]

IterLike: TypeAlias = Union[Iterable, SeriesLike, IndexLike]

GroupKey: TypeAlias = Union[str, int, float, Tuple]

PathType: TypeAlias = Union[pathlib.Path, pathlib.PosixPath]

PathLike: TypeAlias = Union[str, PathType]

# %% ../nbs/03_types.ipynb 15
from typing import (NamedTuple)

# %% ../nbs/03_types.ipynb 17
class TrainValidSplit(NamedTuple):
    train: int
    valid: int

class TrainTestSplit(NamedTuple):
    train: int
    test: int
        
class TrainValidTestSplit(NamedTuple):
    train: int
    valid: int
    test: int

# %% ../nbs/03_types.ipynb 18
SplitTuple2: TypeAlias =  Tuple[Number, Number]
SplitTuple3: TypeAlias =  Tuple[Number, Number, Number]
SplitAsList: TypeAlias =  List[Number]
  
DatasetSplitSpec: TypeAlias = Union[
    TrainValidSplit, TrainTestSplit, TrainValidTestSplit,
    SplitAsList, SplitTuple2, SplitTuple3
]

# %% ../nbs/03_types.ipynb 19
SplitDataset2Idxs: TypeAlias = Tuple[IndexLike, IndexLike]
SplitDataset3Idxs: TypeAlias = Tuple[IndexLike, IndexLike, IndexLike]

SplitDataFrames2: TypeAlias = Tuple[DataFrame, DataFrame]
SplitDataFrames3: TypeAlias = Tuple[DataFrame, DataFrame, DataFrame]

SplitDatasets: TypeAlias = Union[
    SplitDataset2Idxs, SplitDataset3Idxs,
    SplitDataFrames2, SplitDataFrames3,
]

# %% ../nbs/03_types.ipynb 21
from typing import TypeAlias
from matplotlib.colors import Colormap

# %% ../nbs/03_types.ipynb 22
ColorMap: TypeAlias = Colormap

# %% ../nbs/03_types.ipynb 24
from typing import Optional, ClassVar, TypeAlias, Callable
from enum import StrEnum

# %% ../nbs/03_types.ipynb 25
class DirectoryTreeStrings(StrEnum):
    SPACE : ClassVar[str] = '    '
    BRANCH: ClassVar[str] = '│   '    
    TEE   : ClassVar[str] = '├── '
    LAST  : ClassVar[str] = '└── '

# %% ../nbs/03_types.ipynb 26
TreeEntryFunc: TypeAlias = Callable[
    ['PathLike', Optional[str], Optional[str], Optional[str]], str
]

# %% ../nbs/03_types.ipynb 28
from typing import Any, TypeAlias

# %% ../nbs/03_types.ipynb 29
try:
    from rich.tree import Tree as TreeType
    from rich.text import Text as TextType
    from rich.console import Console as ConsoleType
    from rich.progress import Progress as ProgressType
    
except ImportError:
    TreeType = TextType = ConsoleType = ProgressType = Any

RichTree: TypeAlias = TreeType
RichText: TypeAlias = TextType
RichConsole: TypeAlias = ConsoleType
RichProgress: TypeAlias = ProgressType



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_utils.ipynb.

# %% auto 0
__all__ = ['DecompressFunc', 'isiter', 'allinstance', 'allsametype', 'isin', 'arein', 'str_just_alpha', 'str_just_numeric',
           'strip_punc', 'str_split_get', 'str_part_pre', 'str_part_mid', 'str_part_end', 'filter_kwargs_for_func',
           'filter_kwargs_for_class', 'wrangle_kwargs_for_func', 'wrangle_kwargs_for_class', 'get_func_params',
           'copy_to_clipboard', 'is_pathlib', 'is_pathlike', 'is_path', 'str_to_path', 'path_to_str', 'to_abs_expanded',
           'sort_file_first', 'sort_directory_first', 'get_user', 'collapse_user', 'check_ext', 'drop_ext', 'is_tar',
           'is_gz', 'is_targz', 'is_tarball', 'filter_for_gz_files', 'get_gz_files_in_dir', 'decompress_tarball',
           'decompress_gunzip', 'undo_gz', 'RecursiveDecompressor', 'make_missing_dirs', 'dir_dirs',
           'decompress_directory_of_gunzipped_files', 'decompress_tarball_of_gunzipped_files', 'stream_file',
           'download_and_decompress_tarball_of_gunzipped_files', 'make_temp_file', 'urljoin', 'Slice', 'base_init_tree',
           'rich_init_tree', 'init_tree', 'base_entry_fn', 'rich_entry_fn', 'entry_fn', 'walk_dir_tree',
           'DirectoryTree', 'RichDirectory', 'Directory', 'is_matrix', 'undo_npmatrix', 'is_series', 'is_series_like',
           'is_np', 'is_device', 'is_cpu', 'is_mps', 'is_tensor', 'is_torch', 'undo_sparse', 'to_ndarray',
           'AdataExtractor', 'config_exp_logger', 'exp_log_filename', 'exp_param_filename', 'list_exps', 'gen_exp_name',
           'load_exp_params', 'save_exp_params', 'setup_exp', 'is_config_subset', 'find_exps', 'ArchiveDownloader']

# %% ../nbs/02_utils.ipynb 5
import os, inspect, string
from dataclasses import dataclass, field

import numpy as np, pandas as pd

from typing import List, Any, Optional, Callable, Union, Tuple, Iterable, Set, TypeAlias, Type, Dict

# %% ../nbs/02_utils.ipynb 7
def isiter(val: Any) -> bool:    
    return isinstance(val, Iterable)

def allinstance(vals:Any, dtype:Union[Type, TypeAlias]=Any) -> bool:
    return isiter(vals) and all(isinstance(i, dtype) for i in vals)

def allsametype(vals:Any) -> bool:
    if not isiter(vals) or len(vals) == 0: return True
    dtype = type(vals[0])
    return isiter(vals) and all(isinstance(i, dtype) for i in vals)

def isin(val:Any, vals:Iterable) -> bool:
    return val in vals

def arein(vals:Iterable, refs:Iterable) -> bool:                                             
    return isiter(vals) and isiter(refs) and all(isin(v, refs) for v in vals)

# %% ../nbs/02_utils.ipynb 8
def str_just_alpha(s: str) -> str:
    '''Filters a string for just alpha values'''
    return ''.join(list(filter(str.isalpha, s)))

def str_just_numeric(s: str) -> str:
    '''Filters a string for just numeric values'''
    return ''.join(list(filter(str.isnumeric, s)))

def strip_punc(s: str) -> str:
    return s.translate(str.maketrans('', '', string.punctuation))

def str_split_get(s: str, sep: str = '_', idx: int = 0) -> str:
    parts = s.split(sep)
    return parts[idx]    

def str_part_pre(s:str, sep: Optional[str] = ' ') -> str:
    pre, mid, end = s.partition(sep)
    return pre

def str_part_mid(s:str, sep: Optional[str] = ' ') -> str:
    pre, mid, end = s.partition(sep)
    return mid

def str_part_end(s:str, sep: Optional[str] = ' ') -> str:
    pre, mid, end = s.partition(sep)
    return end

# %% ../nbs/02_utils.ipynb 10
def filter_kwargs_for_func(fn: Callable, **kwargs:Optional[dict]):
    params = inspect.signature(fn).parameters
    return {k:v for k,v in kwargs.items() if k in params}

def filter_kwargs_for_class(cls: Callable, **kwargs:Optional[dict]):
    params = inspect.signature(cls.__init__).parameters
    return {k:v for k,v in kwargs.items() if k in params}

def wrangle_kwargs_for_func(
    fn: Callable, 
    defaults: Optional[dict]=None,
    **kwargs:Optional[dict]
) -> dict:
    # copy defaults
    params = (defaults or {}).copy()
    # update with kwargs of our function
    params.update(kwargs or {})
    # filter for only the params that other function accepts
    params = filter_kwargs_for_func(fn, **params)
    return params

def wrangle_kwargs_for_class(
    cls: Callable, 
    defaults: Optional[dict]=None,
    **kwargs:Optional[dict]
) -> dict:
    # copy defaults
    params = (defaults or {}).copy()
    # update with kwargs of our class
    params.update(kwargs or {})
    # filter for only the params that other class accepts
    params = filter_kwargs_for_class(cls, **params)
    return params

def get_func_params(
    fn: Callable, 
    drop_self: Optional[bool] = True,
    drop_before: Optional[int] = 0,
    drop_idxs: Optional[List[int]] = list(),
    drop_names: Optional[List[str]] = list(),
    drop_after: Optional[int] = None,
) -> Dict[str, Any]:
    params = inspect.signature(fn).parameters
    params = {k: v.default for k, v in params.items()}
    if drop_self and 'self' in params: 
        params.pop('self')

    params = {
        n: p for i, (n, p) in enumerate(params.items()) 
        if (
            # is before <= i < after
            (drop_before <= i or (drop_after is not None and i < drop_after))
            # i not in drop_idxs and n not in drop_names
            and (i not in drop_idxs and n not in drop_names)
        )
    }  
    return params

# %% ../nbs/02_utils.ipynb 11
def copy_to_clipboard(text: str):
    command = ""
    
    # macOS and Linux
    if os.name == 'posix':  
        command = 'echo "' + text + '" | pbcopy'
    
    # Windows
    elif os.name == 'nt':  
        command = 'echo ' + text + ' | clip'

    else:
        raise OSError("Unsupported operating system")

    os.system(command)

# %% ../nbs/02_utils.ipynb 13
import os, sys, pwd, atexit, tempfile, inspect, pathlib
import requests, tarfile, gzip, shutil
from dataclasses import dataclass, field
from tqdm.auto import tqdm

from typing import Optional, List, Union, Iterable, Tuple, Any, Callable

# %% ../nbs/02_utils.ipynb 14
from iza.static import (
    EXT_GZ, EXT_TAR, EXT_TAR_GZ,
)

from iza.types import (
    PathType, PathLike
)

# %% ../nbs/02_utils.ipynb 16
def is_pathlib(path:Any) -> bool:
    """Check if the input is a path"""
    return isinstance(path, PathType)

def is_pathlike(path:Any) -> bool:
    """Check if the input is a path"""
    return isinstance(path, (PathLike))

def is_path(path:Any, existance:Optional[bool]) -> bool:
    """Check if the input is a path"""
    exists_q = os.path.exists(os.path.expanduser(path))
    return is_pathlike(path) and (exists_q if existance else True)

def str_to_path(path:str) -> PathType:
    """Convert a string to a pathlib.Path object"""
    return pathlib.Path(path)

def path_to_str(path: PathType) -> str:
    """Convert a pathlib.Path object to a string"""
    return os.fspath(path)

def to_abs_expanded(dirname:Optional[PathLike]=None) -> PathType:
    if dirname is None:
        dirname = os.getcwd()

    dirname = os.path.expanduser(dirname) 
    dirname = os.path.abspath(dirname) 

    if not is_pathlib(dirname):
        dirname = pathlib.Path(dirname)

    return dirname

def sort_file_first(path: PathType):
    return (not path.is_file(), path.name.lower())

def sort_directory_first(path: PathType):
    return (path.is_file(), path.name.lower())


# %% ../nbs/02_utils.ipynb 18
def get_user() -> str:
    user = pwd.getpwuid(os.getuid())[0]
    return user

def collapse_user(path: str) -> str:
    _, rest = path.split(get_user())    
    return '~' + rest

# %% ../nbs/02_utils.ipynb 20
def check_ext(filename:str, extension:str) -> bool:
    if is_pathlib(filename):
        filename = path_to_str(filename)
    has_extension = extension in filename 
    splits = filename.split(extension)
    is_end_of_str = len(splits) >= 2 and splits[-1] == ''
    is_end_of_str = filename.endswith(extension)
    return has_extension and is_end_of_str

def drop_ext(filename:str, extension:Optional[str]=None) -> str:
    if is_pathlib(filename):
        filename = path_to_str(filename)
        
    file = os.path.basename(filename)
    if extension is None:
        file, *_ = file.split('.')
    else:
        file = filename.replace(extension, '')
    return os.path.join(os.path.dirname(filename), file)

# %% ../nbs/02_utils.ipynb 21
def is_tar(filename:str) -> bool:
    return check_ext(filename, EXT_TAR)

def is_gz(filename:str) -> bool:
    return check_ext(filename, EXT_GZ)

def is_targz(filename:str) -> bool:
    return check_ext(filename, EXT_TAR_GZ)

def is_tarball(filename:str) -> bool:
    return is_tar(filename) or is_targz(filename)


# %% ../nbs/02_utils.ipynb 22
def filter_for_gz_files(files:List[str]) -> List[str]:
    return list(filter(lambda f: is_gz(f), files))

def get_gz_files_in_dir(dirname:str) -> List[str]:
    all_files = []

    for (root, dirs, files) in os.walk(dirname):   
        fullpaths = [os.path.join(root, file) for file in files]
        all_files.extend(fullpaths)
    
    gz_files = filter_for_gz_files(all_files)
    return gz_files

# %% ../nbs/02_utils.ipynb 24
def decompress_tarball(filename:str, remove:bool=False) -> Tuple[str, Optional[EOFError]]:
    '''
    Returns
    -------
        dirname : str
            The name of the archive e.g. `~/Downloads/fluentbio.tar.gz` would
            yield `~/Downloads/fluentbio`


    Notes
    -----
    FluentBio has a weird gzip so it complains when it is 
        actually fine
    '''
    error = None
    decompress_dir = os.path.dirname(filename)
    ext = EXT_TAR_GZ if is_targz(filename) else EXT_TAR if is_tar(filename) else EXT_GZ

    dirname = drop_ext(filename, ext)
    try:
        with tarfile.open(filename) as tarball:
            tarball.extractall(decompress_dir)
            tarball.close()

    except EOFError as error:
        pass
    
    if os.path.isdir(dirname) and remove:
        os.remove(filename)

    return dirname, error


def decompress_gunzip(filename:str, remove:bool=False) -> Tuple[str, Optional[EOFError]]:
    '''
    Returns
    -------
        file : str
            The name of the decompressed file e.g. `~/Downloads/fluentbio.tsv.gz` would
            yield `~/Downloads/fluentbio.tsv`

    Notes
    -----
    FluentBio has a weird gzip so it complains when it is 
        actually fine
    '''
    error = None
    decompressed_file = drop_ext(filename, EXT_GZ)
    try:             
        with gzip.open(filename, 'rb') as gunzipped:
            with open(decompressed_file, 'wb') as unzipped:
                shutil.copyfileobj(gunzipped, unzipped)     
                   
    except EOFError as error:
        pass

    if os.path.isfile(decompressed_file) and remove:
        os.remove(filename)

    return decompressed_file, error

def undo_gz(filename: str, remove:bool=False) -> str:
    if is_targz(filename):
        filename, _ = decompress_tarball(filename, remove)
    elif is_gz(filename):
        filename, _ = decompress_gunzip(filename, remove)
    elif is_tar(filename):
        pass
    elif is_tarball(filename):
        filename, _ = decompress_tarball(filename, remove)
    return filename

# %% ../nbs/02_utils.ipynb 25
DecompressFunc = Callable[[str, bool], str]

@dataclass
class RecursiveDecompressor:
    dirname: PathType
    entries: Optional[List[str]] = field(default_factory=list, repr=False)
    strategy: Optional[DecompressFunc] = field(default=undo_gz, repr=False)
    remove: Optional[bool] = field(default=True, repr=False)
    progress: Optional[Any] = field(default=None, repr=False, init=True)

    total: Optional[int] = field(default=None, repr=False, init=False)
    done: Optional[int] = field(default=None, repr=False, init=False)
    task: Optional[Any] = field(default=None, repr=False, init=False)

    def __post_init__(self):
        if not is_pathlib(self.dirname): 
            self.dirname = pathlib.Path(self.dirname)
            
        if self.entries is None: 
            self.entries = os.listdir(self.dirname)

        self.done = 0
        self.total = len(self.entries)

    def execute(self):
        return self.decompress()

    def update_total(self, increment: int):
        self.total += increment
        if self.progress is not None:
            self.progress.update(self.task, total=self.total)
        else:
            self.tqdm_bar.total = self.total
            self.tqdm_bar.refresh()

    def decrease_total(self):
        self.total -= 1
        if self.progress is not None:
            self.progress.update(self.task, total=self.total)
        else:
            self.tqdm_bar.total = self.total
            self.tqdm_bar.refresh()

    def decompress(self):
        files = [self.dirname / entry for entry in self.entries]

        if self.progress is not None:
            with self.progress:
                self.task = self.progress.add_task("[cyan]Extracting...", total=self.total)
                for file in files:
                    self.recurse(file)
                    self.progress.advance(self.task)
        else:
            with tqdm(total=self.total, desc='Extracting') as self.tqdm_bar:
                for file in files:
                    self.recurse(file)
                    self.tqdm_bar.update()

    def recurse(self, current:PathLike):
        filename = path_to_str(current)
        nextfile = self.strategy(filename, self.remove)
        self.done += 1
        nextfile = pathlib.Path(nextfile)
        
        # If the decompressed file is a directory, recursively decompress it
        if nextfile.is_dir():
            for root, dirs, files in os.walk(nextfile):
                for f in files:
                    self.update_total(1)
                    self.recurse(pathlib.Path(root) / f)
                    
        elif self.remove:
            if filename in self.entries:
                self.entries.remove(filename)
            self.decrease_total()


# %% ../nbs/02_utils.ipynb 27
def make_missing_dirs(dirs:List[str]):
    if isinstance(dirs, str):
        dirs = [dirs]
        
    elif is_pathlib(dirs):
        dirs = [dirs]
        
    for d in dirs:
        if not os.path.exists(d):
            os.makedirs(d)
            
def dir_dirs(dirname:str) -> List[str]:
    entries = os.listdir(dirname)
    is_subdir = lambda e : os.path.isdir(os.path.join(dirname, e))
    return list(filter(is_subdir, entries))

# %% ../nbs/02_utils.ipynb 28
def decompress_directory_of_gunzipped_files(
    dirname:str, desc:Optional[str]=None, remove:Optional[bool]=False
) -> None:
    if desc is None:
        desc = dirname.split('/')[-1]

    gz_files = get_gz_files_in_dir(dirname)
    for filename in tqdm(gz_files, desc=desc):
        decomp_filename, error = decompress_gunzip(filename, remove)   


def decompress_tarball_of_gunzipped_files(
    filename:str, desc:Optional[str]=None, remove:Optional[bool]=False
) -> None:
    # NOTE: initial decompress of .tar.gz
    dirname, error = decompress_tarball(filename)

    if desc is None:
        desc = dirname.split('/')[-1]

    # NOTE: decompress all internal .gz files
    decompress_directory_of_gunzipped_files(dirname, desc, remove)

# %% ../nbs/02_utils.ipynb 30
def stream_file(uri:str, filename:Optional[str]=None, desc:Optional[str]=None) -> None:
    '''
    Parameters
    ----------
    uri : str
        The URI to download

    filename : str, optional
        The fullpath name of the file to download. Defaults to 
        `~/Downloads/os.path.basename(uri)`.

    desc : str, optional
        The description of the `tqdm` progress bar. Defaults to 
        `os.path.basename(uri)`.
    '''
    if filename is None:
        download_dir = os.path.expanduser(f'~/Downloads')        
        filename = os.path.join(download_dir, os.path.basename(uri))

    basename = os.path.basename(filename)
    if desc is None:
        desc = basename

    response = requests.get(uri, stream=True)
    total = int(response.headers.get('content-length', 0))
    
    with tqdm.wrapattr(
        open(filename, 'wb'), 'write', 
        miniters=1, desc=desc, total=total
    ) as fout:
        for chunk in response.iter_content(chunk_size=4096):
            fout.write(chunk)

# %% ../nbs/02_utils.ipynb 32
def download_and_decompress_tarball_of_gunzipped_files(
    uri:str, download_dir:str=None, desc:Optional[str]=None, remove:Optional[bool]=False
):
    filename = os.path.basename(uri)
    fullpath = os.path.join(download_dir, filename)

    if desc is None:
        description = f'Downloading {filename}'


    # NOTE: Amazon --> filtered_matrix.tar.gz
    stream_file(uri, description)

    if desc is None:
        description = f'Decompressing {filename}'
    # NOTE: filtered_matrix.tar.gz --> filtered_matrix/**/file.tsv
    decompress_tarball_of_gunzipped_files(fullpath, desc, remove)


# %% ../nbs/02_utils.ipynb 34
def make_temp_file(**kwargs: Any) -> tempfile.NamedTemporaryFile:
    temp = tempfile.NamedTemporaryFile(**kwargs)
    @atexit.register
    def delete_temp() -> None:
        temp.close()
    return temp

# %% ../nbs/02_utils.ipynb 36
from urllib3.util.url import parse_url

# %% ../nbs/02_utils.ipynb 37
def urljoin(*parts: str) -> str:
    return parse_url('/'.join(s.strip('/') for s in parts)).url

# %% ../nbs/02_utils.ipynb 39
from dataclasses import dataclass, field
import numpy as np, pandas as pd

from typing import List, Union, Tuple

# %% ../nbs/02_utils.ipynb 42
@dataclass
class Slice:
    """A class for representing a slice and providing conversion to other formats."""
    slc: slice = field(default_factory=slice)

    @property
    def start(self):
        try:
            return self._start
        except AttributeError:
            self._start = self.slc.start
        return self._start
    
    @start.setter
    def start(self, value):
        """Sets the start index."""
        if value < 0:
            raise ValueError("Slice indices must be non-negative.")
        self._start = value

    @property
    def stop(self):
        try:
            return self._stop
        except AttributeError:
            self._stop = self.slc.stop
        return self._stop
    
    @stop.setter
    def stop(self, value):
        """Sets the stop index."""
        if value < 0:
            raise ValueError("Slice indices must be non-negative.")
        self._stop = value
    
    @property
    def step(self):
        """Gets the step index."""
        try:
            return self._step
        except AttributeError:
            self._step = self.slc.step
        return self._step
    
    @step.setter
    def step(self, value):
        """Sets the step index."""
        if value < 0:
            raise ValueError("Slice step must be non-negative.")
        self._step = value
    
    def __post_init__(self):
        if self.start is None:
            self.start = 0

        if self.stop is None:
            self.stop = min(0, self.start, max(1, self.start))

        if self.step is None:
            self.step = 1

    def totuple(self) -> Tuple[Union[int, float, None], Union[int, float, None], Union[int, float, None]]:
        """Converts the slice to a tuple."""
        return (self.start, self.stop, self.step)
    
    def toslice(self) -> slice:
        """Converts the updated slice."""
        return slice(self.start, self.stop, self.step)
    
    def tolist(self) -> List[Union[int, float]]:
        """Converts the slice to a list."""
        return list(range(self.start, self.stop, self.step))
    
    def todict(self) -> List[Union[int, float]]:
        """Converts the slice to a dict."""
        return dict(zip('start stop step'.split(), self.totuple()))
        

    def astype(self, dtype:str):
        """Converts the slice to a specified format."""
        if dtype in {'list', list}:
            return self.tolist()
        elif dtype in {'numpy', np.ndarray}:
            return np.array(self.tolist())
        elif dtype in {'pandas', pd.Series}:
            return pd.Series(self.tolist())
        elif dtype in {'tuple', tuple}:
            return self.totuple()
        elif dtype in {'dict', dict}:
            return self.todict()
        elif dtype in {'slice', slice}:
            return self.toslice()
        return self

# %% ../nbs/02_utils.ipynb 44
import os, pathlib 
from pathlib import Path
from dataclasses import dataclass, field, KW_ONLY
from typing import Optional, List, ClassVar, Any, TypeAlias, Callable
from enum import StrEnum

# %% ../nbs/02_utils.ipynb 45
from iza.types import (
    PathLike, PathType,
    DirectoryTreeStrings, TreeEntryFunc,
    RichTree, RichText, RichConsole, RichProgress
)
from .static import EXT_PY

#| export
from .imp import RichImp
from ipos.imp import is_mod, is_var_imp, Module, Imp

# %% ../nbs/02_utils.ipynb 46
try:    
    from rich.tree import Tree
    from rich.text import Text    
    from rich.filesize import decimal
    from rich import get_console
    from rich.console import Console
    from rich.progress import Progress

except ImportError:
    Tree = None
    Text = None 
    Progress = None

# %% ../nbs/02_utils.ipynb 50
def base_init_tree(dirname: str):
    return None

def rich_init_tree(dirname):
    return Tree(f'[link file://{dirname}]{dirname}', guide_style='bold bright_blue')

def init_tree(dirname):
    try:
        return rich_init_tree(dirname)
    except:
        return base_init_tree(dirname)

# %% ../nbs/02_utils.ipynb 52
def base_entry_fn(path:Path, prefix:str='', pointer:str='', suffix:str='') -> str:
    name = path.name
    return f'{prefix}{pointer}{name}{suffix}'

def rich_entry_fn(path:Path, prefix:str='', pointer:str='', suffix:str='') -> str:
    text = Text(path.name)
    size = decimal(path.stat().st_size)
    text.stylize(f'link file://{path}')
    text.append(f' ({size})', 'cyan')
    return text

def entry_fn(path:Path, prefix:str='', pointer:str='', suffix:str='') -> str:
    try:
        return rich_entry_fn(path, prefix, pointer, suffix)
    except:
        return base_entry_fn(path, prefix, pointer, suffix)


# %% ../nbs/02_utils.ipynb 54
def walk_dir_tree(
    dirname: PathLike, 
    prefix: str = '',
    hidden: Optional[bool] = False, 
    tree: Optional[RichTree] = None,
    entry_fn: TreeEntryFunc = entry_fn,
):    
    '''
    A recursive generator, given a directory Path object
    will yield a visual tree structure line by line
    with each line prefixed by the same characters
    Notes
    -----
    Adapted from https://stackoverflow.com/a/59109706/5623899
    '''
    # NOTE: sort_directory_first, to_abs_expanded, rich_file defined in _02_utils/_08_modules.ipynb
    dirname = Path(to_abs_expanded(dirname))
    contents = sorted(Path(dirname).iterdir(), key=sort_directory_first)
    
    SPACE, BRANCH, TEE, LAST = DirectoryTreeStrings
    pointers = [TEE] * (len(contents) - 1) + [LAST]
    
    for pointer, path in zip(pointers, contents):
        # Remove hidden files
        if path.name.startswith('.') and not hidden:
            continue
        
        name = entry_fn(path, prefix, pointer, suffix='')
        yield name
        
        branch = None
        if tree is not None:
            branch = tree.add(entry_fn(path))
    
        if path.is_dir():
            # NOTE: space because last, └── , above so no more |
            extension = BRANCH if pointer == TEE else SPACE
            yield from walk_dir_tree(path, prefix=f'{prefix}{extension}', hidden=hidden, tree=branch, entry_fn=entry_fn)

# %% ../nbs/02_utils.ipynb 56
@dataclass
class DirectoryTree:
    dirname: str
    
    hidden: Optional[bool] = False
    entry_fn: Optional[TreeEntryFunc] = field(default=entry_fn)
    tree: Optional[RichTree] = field(init=False, default=None)

    coerce_str: ClassVar[Optional[bool]] = field(init=False, default=True, repr=False)
    
    def __post_init__(self):
        # NOTE: defined in _02_utils/_01_files.ipynb
        self.dirname = Path(to_abs_expanded(self.dirname))
        if self.tree is None:
            self.tree = init_tree(self.dirname)

    def tree_generator(self):
        yield from walk_dir_tree(
            self.dirname, hidden=self.hidden, 
            tree=self.tree, entry_fn=self.entry_fn
        )
        
    def get_tree_lines(self) -> List[str]:        
        tree_gen = self.tree_generator()
        lines = [
            str(line) if self.coerce_str else line 
            for line in tree_gen
        ]
        return lines
    
    def make_tree_str(self) -> str:        
        lines = self.get_tree_lines()
        tree_str = '\n'.join([str(self.dirname), *lines])
        return tree_str
   
    def print(self) -> None:        
        tree_str = self.make_tree_str()
        print(tree_str)
        return

    def __repr__(self):        
        tree_str = self.make_tree_str()
        return tree_str  

# %% ../nbs/02_utils.ipynb 57
@dataclass
class RichDirectory(DirectoryTree):
    _: KW_ONLY
    console: Optional[RichConsole] = field(default_factory=Console, init=True, repr=False)    
    _imp: Imp = field(default_factory=RichImp, init=False, repr=False)
    
    coerce_str: ClassVar[Optional[bool]] = field(init=False, default=False, repr=False)

    def __post_init__(self):
        super().__post_init__()
        self._imp = RichImp()
    
        if self.console is None and is_mod(self._imp._module):
            self.console = get_console()

    def print_rich(self) -> None:    
        lines = self.get_tree_lines(_coerce_str=False)
        self.console.print(self.tree)

# %% ../nbs/02_utils.ipynb 58
@dataclass
class Directory(DirectoryTree):
    dirname: PathLike
    pass

# %% ../nbs/02_utils.ipynb 60
import os
from pathlib import Path
from dataclasses import dataclass, field, KW_ONLY
from typing import Optional, List, ClassVar

# %% ../nbs/02_utils.ipynb 62
import pandas as pd

# %% ../nbs/02_utils.ipynb 63
from iza.static import (
    ADATA, MATRIX, BARCODES, FEATURES, EXT_H5, EXT_MTX, EXT_TSV,
    GENE_SYMBOL, ENSEMBL_ID
)
from iza.types import (
    AnnData
)
from tqdm.auto import tqdm

# %% ../nbs/02_utils.ipynb 64
try: 
    import scanpy as sc, scprep

    # NOTE: Directory defined in _02_utils/_02_directory.ipynb
    @dataclass
    class FilterMatrixDirectory(Directory):
        _: KW_ONLY
        ADATA_FILE: ClassVar[str] = f'{ADATA}{EXT_H5}'
        MATRIX_FILE: ClassVar[str] = f'{MATRIX}{EXT_MTX}'
        BARCODES_FILE: ClassVar[str] = f'{BARCODES}{EXT_TSV}'
        FEATURES_FILE: ClassVar[str] = f'{FEATURES}{EXT_TSV}'
        

        def __post_init__(self):    
            try:
                if not self.has_adata:
                    self.make_adata()
            except Exception as err:
                raise err

        def __repr__(self):
            base = os.path.basename(self.dirname)
            srep = f'FilteredMatrix(valid: {self.is_valid}, adata: {self.has_adata})'        
            srep += '\n'
            srep += super(FilterMatrixDirectory, self).__repr__()
            return srep
                    
        @property
        def adata_filename(self) -> str:
            return os.path.join(self.dirname, self.ADATA_FILE)

        @property
        def matrix_filename(self) -> str:
            return os.path.join(self.dirname, self.MATRIX_FILE)
        
        @property
        def barcodes_filename(self) -> str:
            return os.path.join(self.dirname, self.BARCODES_FILE)
        
        @property
        def features_filename(self) -> str:
            return os.path.join(self.dirname, self.FEATURES_FILE)

        @property
        def has_adata(self) -> bool:
            return os.path.isfile(self.adata_filename)

        @property
        def has_matrix(self) -> bool:
            return os.path.isfile(self.matrix_filename)

        @property
        def has_barcodes(self) -> bool:
            return os.path.isfile(self.barcodes_filename)

        @property
        def has_features(self) -> bool:
            return os.path.isfile(self.features_filename)

        @property
        def is_valid(self) -> bool:
            return all([self.has_matrix, self.has_barcodes, self.has_features])

        def make_adata(self) -> AnnData:
            if self.has_adata:
                return

            steps = (FEATURES, BARCODES, MATRIX, 'combine', ADATA)
            
            desc = os.path.basename(self.dirname)

            steps = tqdm(steps, desc=desc, leave=True)        
            for step in steps:
                steps.set_postfix(stage=step)
                match step:
                    case 'features':
                        features = pd.read_csv(self.features_filename, sep='\t', header=None)
                        features.columns = [ENSEMBL_ID, GENE_SYMBOL, 'feature_type']
                        features.index = pd.Series(features.ensembl_id.copy().values)

                    case 'barcodes':
                        barcodes = pd.read_csv(self.barcodes_filename, sep='\t', header=None)
                        barcodes.columns = [BARCODES]
                        barcodes.index = pd.Series(barcodes.barcodes.copy().values)

                    case 'matrix':
                        matrix = scprep.io.load_mtx(self.matrix_filename, sparse=True).T

                    case 'combine':
                        data = pd.DataFrame.sparse.from_spmatrix(
                            matrix, columns=features.index, index = barcodes.index
                        )
                        del matrix

                    case 'adata':
                        adata = sc.AnnData(X=data.values, obs=barcodes, var=features, dtype='float32')
                        adata.write(self.adata_filename)
                        
                    case _:
                        pass

            return adata

        def get_adata(self) -> AnnData:
            adata = sc.read_h5ad(self.adata_filename)
            return adata

except ImportError as err:
    @dataclass
    class FilterMatrixDirectory(Directory):
        _: KW_ONLY
        ADATA_FILE: ClassVar[str] = f'{ADATA}{EXT_H5}'
        MATRIX_FILE: ClassVar[str] = f'{MATRIX}{EXT_MTX}'
        BARCODES_FILE: ClassVar[str] = f'{BARCODES}{EXT_TSV}'
        FEATURES_FILE: ClassVar[str] = f'{FEATURES}{EXT_TSV}'
        

        def __post_init__(self):    
            raise ImportError('FilterMatrixDirectory requires scprep and scanpy to be installed')
        
    pass

# %% ../nbs/02_utils.ipynb 66
import os
import numpy as np, pandas as pd
from typing import Optional, List, ClassVar, Any

# %% ../nbs/02_utils.ipynb 67
from .types import Tensor, Device, SeriesLike, ndarray

# %% ../nbs/02_utils.ipynb 70
def is_matrix(arr: SeriesLike) -> bool:
    '''
    Checks whether or not `arr` is a np.matrix
    
    Parameters
    ----------    
    arr : SeriesLike
        object to check whether or not it is a np.matrix.
    
    Returns
    -------
    result : bool
    '''
    return isinstance(arr, np.matrix)

def undo_npmatrix(arr: SeriesLike) -> SeriesLike:  
    '''
    Given a tensor converts it to a numpy array
    
    Parameters
    ----------    
    tensor : Tensor
        
    
    Returns
    -------
    arr : ndarray
    
    Notes
    -----
    - several graphtool functions use dependencies which rely on
        the deprecated numpy class `np.matrix`.
        
    - these functions appear to be related to scipy sparse linalg
        methods.
    '''
    if is_matrix(arr):
        return np.array(arr)
    return arr

def is_series(arr: SeriesLike) -> bool:
    '''
    Checks whether or not `arr` is a pd.Series
    
    Parameters
    ----------    
    arr : SeriesLike
        object to check whether or not it is a pd.Series.
    
    Returns
    -------
    result : bool
    '''
    return isinstance(arr, pd.Series)

def is_series_like(series_q: SeriesLike) -> bool:
    '''
    Checks whether or not `series_q` is SeriesLike
    i.e. something that is probably data.
    
    Parameters
    ----------    
    series_q : SeriesLike
        object to check whether or not it is a SeriesLike.
    
    Returns
    -------
    result : bool
    '''
    return isinstance(series_q, SeriesLike)

def is_np(arr_q: SeriesLike) -> bool:
    '''
    Checks whether or not `arr_q` is a ndarray
    
    Parameters
    ----------    
    arr_q : SeriesLike
        object to check whether or not it is a SeriesLike.
    
    Returns
    -------
    result : bool
    '''
    return isinstance(arr_q, ndarray)


# %% ../nbs/02_utils.ipynb 71
def is_device(device_q: Device) -> bool:
    '''
    Checks whether or not `device_q` is a valid
    pytorch device type.
    
    Parameters
    ----------    
    device_q : Device
        object to check whether or not it is a pytorch device.
    
    Returns
    -------
    result : bool        
    
    Notes
    -----
    - There is an execption. `NoneType` is a valid
        pytorch device. Here we return `False` instead.
    '''
    if device_q is None:
        return False
    return isinstance(device_q, Device)

def is_cpu(tensor: Tensor) -> bool:
    '''
    Checks whether or not `tensor` is on cpu
    
    Parameters
    ----------    
    tensor : Tensor
        object to check whether or not it is on cpu.
    
    Returns
    -------
    result : bool
    '''
    # assert is_tensor(tensor)
    if not hasattr(tensor, 'device'):
        return True
    return tensor.device.type == 'cpu'

def is_mps(tensor: Tensor) -> bool:
    '''
    Checks whether or not `tensor` is on cpu
    
    Parameters
    ----------    
    tensor : Tensor
        object to check whether or not it is on cpu.
    
    Returns
    -------
    result : bool
    '''
    # assert is_tensor(tensor)
    if not hasattr(tensor, 'device'):
        return False
    return tensor.device.type == 'mps'

def is_tensor(tensor_q: SeriesLike) -> bool:
    '''
    Checks whether or not `tensor_q` is a pytorch tensor
    
    Parameters
    ----------    
    tensor_q : Tensor
        object to check whether or not it is a pytorch tensor
    
    Returns
    -------
    result : bool
    '''
    return isinstance(tensor_q, Tensor)


def is_torch(tensor_q: SeriesLike) -> bool:
    '''
    Alias for `is_tensor`.
    
    Parameters
    ----------    
    tensor_q : Tensor
        object to check whether or not it is a pytorch tensor
    
    Returns
    -------
    result : bool
    
    See Also
    --------
    is_tensor
    '''
    return is_tensor(tensor_q)

# %% ../nbs/02_utils.ipynb 72
def undo_sparse(arr: SeriesLike) -> SeriesLike:  
    '''
    Given a arr tries to make it a dense array
    
    Parameters
    ----------    
    arr : SeriesLike
        
    
    Returns
    -------
    arr : ndarray
    '''
    if hasattr(arr, 'toarray'):
        arr = arr.toarray()

    if hasattr(arr, 'todense'):
        arr = arr.todense()

    return arr

def to_ndarray(arr):
    if is_np(arr):
        return arr
        
    arr = undo_npmatrix(arr)
    
    arr = undo_sparse(arr)

    if not is_np(arr):
        arr = np.array(arr)

    return arr

# %% ../nbs/02_utils.ipynb 74
import os, random
import numpy as np

from dataclasses import dataclass, field, KW_ONLY
from typing import Optional, List, ClassVar, Any

# %% ../nbs/02_utils.ipynb 75
from .types import Tensor, Device, SeriesLike, ndarray

# %% ../nbs/02_utils.ipynb 77
try:
    import torch
    def ensure_device(device: Device) -> Device:
        '''
        Given a valid device type attempts to instantiant 
        a pytorch device object i.e. `device='cpu'` will
        return `torch.device('cpu')`.
        
        Parameters
        ----------    
        device : Device
            a valid pytorch device type, possible a string.
        
        Returns
        -------
        device : torch.device        
        
        Raises
        ------
        RuntimeError
            same error if `torch.device(device)` fails
        '''
        if device is None:
            return device    
        try:
            return torch.device(device)
        except RuntimeError as err:
            raise err
        return device
    
    def to_cuda(tensor: Tensor) -> Tensor:
        '''
        Given a tensor, ensures that it is on cuda.
        
        Parameters
        ----------    
        tensor : Tensor
            
        
        Returns
        -------
        tensor : Tensor
        '''
        return tensor.cuda()

    def to_mps(tensor: Tensor) -> Tensor:
        '''
        Given a tensor, ensures that it is on mac silicon.
        
        Parameters
        ----------    
        tensor : Tensor
            
        
        Returns
        -------
        tensor : Tensor
        '''
        return tensor.to(torch.device('mps'))
    
    
    def to_torch(
        arr: SeriesLike,
        cuda: Optional[bool] = False,
        mps: Optional[bool] = False,
        device: Optional[Device] = None,
        dtype: Optional[Any] = None
    ) -> Tensor:
        '''
        Given data, ensures that it is a pytorch Tensor.
        
        Parameters
        ----------    
        arr : SeriesLike
        
        cuda : bool, default=False
            whether to return the tensor on cuda
            
        mps : bool, default=False
            whether to return the tensor on mps
            
        device : Device, optional
            whether to return the tensor on given device
            
        Returns
        -------
        tensor : Tensor
            the input array as a pytorch tensor
            
        Notes
        -----
        - `device` takes priority over `cuda` and `mps`
        '''
        tensor = torch.as_tensor(arr)
        if device is not None:
            tensor = tensor.to(device)
        elif cuda:
            tensor = to_cuda(tensor)
        elif mps:
            tensor = to_mps(tensor)    
        
        if dtype is not None:
            dtype = coerce_mps_dtype(dtype, tensor.device, assume_on_mps=False)
            tensor = tensor.to(dtype)

        return tensor
    
    #| export
    def to_np(tensor:Tensor) -> ndarray:
        '''
        Given a tensor converts it to a numpy array
        
        Parameters
        ----------    
        tensor : Tensor
            
        
        Returns
        -------
        arr : ndarray
        '''
        assert is_tensor(tensor)
        if not hasattr(tensor, 'detach'):
            try:
                return np.array(tensor)
            except Exception as err:
                raise err
        try:
            return tensor.detach().clone().cpu().numpy()
        except Exception as err:
            raise err
    

    def is_mps_available() -> bool:
        '''
        Checks whether or not pytorch has mps availble (version) and was built with mps in mind.

        Returns
        -------
        result : bool
        '''
        maybe_mps = torch.backends.mps.is_available()
        built_mps = torch.backends.mps.is_built()
        return maybe_mps and built_mps
    
    def coerce_mps_dtype(
        dtype, 
        device: Optional[Device] = None, 
        assume_on_mps: Optional[bool] = True
    ):
        '''
        Makes sure `tensor` is `torch.float32` if `tensor.dtype` is `torch.float64`
        if `tensor.device` is `'mps'`.
        
        Parameters
        ----------    
        dtype : any
            dtype to check against
        
        device : Device, default=None
            the device of the tensor or model from which the `dtype` comes from. If provided
            will be used to detemine whether or not to make `torch.float64`, `torch.float32`
            only if the device is actually `'mps'`.

        assume_on_mps: bool, default=True
            whether or not to assume that the device of choice is `'mps'`. Setting this to
            `True` will result in `dtype` of `torch.float64` being converted to `torch.float32`
            to try and silently fix mps errors

        Returns
        -------
        dtype : any
            the dtype, corrected for mps if needed
        '''
        could_be_mps = is_mps_available()
        
        is_float64 = dtype == torch.float64

        if device is not None:
            is_device_mps = device.type == 'mps'
            if is_device_mps:
                assume_on_mps = True

            elif device.type == 'cuda':
                assume_on_mps = False

        
        # NOTE: float64 not availble on mps, coerce to float32
        # NOTE: could_be_mps and assume_on_mps both needed as
        #       device might not be provided.
        if could_be_mps and assume_on_mps and is_float64:
            return torch.float32
        
        return dtype
    
    def ensure_mps_dtype(tensor: Tensor) -> Tensor:
        '''
        Makes sure `tensor` is `torch.float32` if `tensor.dtype` is `torch.float64`
        if `tensor.device` is `'mps'`.
        
        Parameters
        ----------    
        tensor : Tensor
            pytorch tensor to maybe change dtype of
        
        Returns
        -------
        tensor : Tensor
        '''
        dtype = tensor.dtype

        # NOTE: we don't assume mps as we explicitly pass the device
        dtype = coerce_mps_dtype(dtype, tensor.device, assume_on_mps=False)

        tensor = tensor.to(dtype)
        return tensor

    def move_to(
        tensor: Tensor, other: Tensor, 
        dtype: Optional[Any] = None, do_dtype: Optional[bool] = True
    ) -> Tensor:
        '''
        Makes sure `tensor` is on the same device as `other`
        
        Parameters
        ----------    
        tensor : Tensor
            pytorch tensor to change device of
            
        other : Tensor
            pytorch tensor we want `tensor` to be on
            
        dtype : optional
            the data type to make `tensor`. If `None` will infer it
            from `other`
            
        do_dype: bool, default=True
            whether or not to just match the device of `other` or also
            the dtype
        
        Returns
        -------
        tensor : Tensor
        '''
        
        if not is_tensor(tensor):
            tensor = to_torch(tensor)
            
        # NOTE: dtype not provided, so we will infer it
        if dtype is None:
            # NOTE: this little line solves mps float64 issues since 
            #       infer our tensor types and move them accordingly
            other = ensure_mps_dtype(other)
            dtype = other.dtype

        if do_dtype:
            tensor = tensor.to(dtype)

        tensor = tensor.to(other.device)
        
        return tensor

except ImportError as err:
    identity = lambda x: x
    ensure_device = identity
    to_cuda = identity
    to_mps = identity
    to_torch = lambda arr, cuda, mps, device, dtype: arr
    to_np = identity
    is_mps_available = lambda: False
    coerce_mps_dtype = lambda dtype, device, assume_on_mps: dtype
    ensure_mps_dtype = identity
    move_to = lambda tensor, other, dtype, do_dtype: tensor
    pass



# %% ../nbs/02_utils.ipynb 78
try:
    import torch, pytorch_lightning as pl
    def set_seeds(seed: int) -> None:
        '''
        Calls a bunch of seed functions with `seed`
        
        Parameters
        ----------
        seed : int
        '''    
        torch.manual_seed(seed)
        random.seed(seed)
        np.random.seed(seed)    
        pl.seed_everything(seed)
except ImportError as err:
     def set_seeds(seed: int) -> None:
         random.seed(seed)
         np.random.seed(seed)

# %% ../nbs/02_utils.ipynb 80
from dataclasses import dataclass, field
import numpy as np, pandas as pd

from typing import List, Any, Optional

# %% ../nbs/02_utils.ipynb 81
from .types import AnnData, ndarray, DataFrame
from .static import X_MAGIC, PHATE, X_PHATE

# %% ../nbs/02_utils.ipynb 83
@dataclass
class AdataExtractor:
    adata: AnnData
    layer: Optional[str] = X_MAGIC
    x_emb: Optional[str] = X_PHATE

    dim_str: Optional[str] = None
    use_hvg: Optional[bool] = True

    @property
    def has_hvg(self):
        return hasattr(self.adata, 'var') and hasattr(self.adata.var, 'highly_variable')
    
    @property
    def has_emb(self):
        return hasattr(self.adata, 'obsm') and self.x_emb in self.adata.obsm.keys()

    def get_layer(self) -> ndarray:
        layer = self.sdata().layers.get(self.layer, None)

        if layer is None:
            layer = self.sdata().X

        if hasattr(layer, 'toarray'):
            layer = layer.toarray()

        if hasattr(layer, 'todense'):
            layer = layer.todense()

        return layer
    
    def get_emb(self) -> ndarray:
        emb = self.sdata().obsm.get(self.x_emb, None)
        if emb is None:
            raise ValueError(f'No embedding found in adata.obsm {self.sdata().obsm.keys()}')

        # NOTE: defined in _02_utils/_05_guards.ipynb
        emb = to_ndarray(emb)
        return emb

    @property
    def axis_str(self):
        if self.dim_str:
            return self.dim_str
        return self.x_emb.replace('X_', '').upper()
    
    @property
    def emb_cols(self):
        ndim = self.get_emb().shape[1]
        cols = [f'{self.axis_str}_{i+1}' for i in range(ndim)]
        return cols
        
    def sdata(self):
        if self.use_hvg and self.has_hvg:
            return self.adata[:, self.adata.var.highly_variable]
        return self.adata
    
    def get_df_cnt(self) -> DataFrame:
        layer = self.get_layer()

        cols = self.sdata().var.index
        idxs = self.sdata().obs.index
        df = pd.DataFrame(layer, index=idxs, columns=cols)
        return df
    
    def get_df_emb(self) -> DataFrame:
        emb = self.get_emb()
        
        cols = self.emb_cols
        idxs = self.sdata().obs.index
        df = pd.DataFrame(emb, index=idxs, columns=cols)
        return df
    
    @property
    def df_cnt(self):
        return self.get_df_cnt()
    
    @property
    def df_emb(self):
        return self.get_df_emb()

# %% ../nbs/02_utils.ipynb 85
import os, yaml, datetime, logging

# %% ../nbs/02_utils.ipynb 86
def config_exp_logger(path):
    '''
    Arguments:
    ----------
        path (str): full path to experiment, i.e. 
            `<path-to-experiments-dir>/<experiment-timestamp>`
    Returns:
    ----------
        logger
    '''
    basename = os.path.basename(path)
    logger = logging.getLogger(basename)

    logging.basicConfig(
        filename=exp_log_filename(path), 
        encoding='utf-8',
        level=logging.DEBUG,
        format='%(asctime)s\t%(levelname)s:%(message)s',
        datefmt='%d/%m/%Y %I:%M:%S %p',
        filemode='w'
    )
    logger.info(f'Experiment path created {path}')
    return logger

def exp_log_filename(path):
    '''
    Arguments:
    ----------
        path (str): full path to experiment, i.e. 
            `<path-to-experiments-dir>/<experiment-timestamp>`
    Returns:
    ----------
        log_file (str): full path to provided experiment's log file
    '''
    return os.path.join(path, 'log.txt')

def exp_param_filename(path):
    '''
    Arguments:
    ----------
        path (str): full path to experiment, i.e. 
            `<path-to-experiments-dir>/<experiment-timestamp>`
    Returns:
    ----------
        param_file (str): full path to provided experiment's parameter file
    '''
    return os.path.join(path, 'params.yml')

def list_exps(path):
    '''
    Notes:
    ----------
        - an experiment is defined as a directory containing a `'params.yml'` file.
    Arguments:
    ----------
        path (str): full path to experiments directory, i.e. 
            `<path-to-experiments-dir>`
    Returns:
    ----------
        experiments (str[]): experiments (subdirectories) in the specified 
            `path`.
    '''
    test_fn = lambda el: os.path.isdir(el) and os.path.isfile(exp_param_filename(el))
    return list(filter(test_fn, os.listdir(path)))

def gen_exp_name(name=None):
    '''    
    Returns:
    ----------
        exp_name (str): timestamp to serve as experiment name
    '''
    if name is None:
        now =  datetime.datetime.now()
        return now.strftime("%Y_%m_%d-%I_%M_%S_%p")
    return name
        
def load_exp_params(path):
    '''
    Arguments:
    ----------
        path (str): full path to experiment, i.e. 
            `<path-to-experiments-dir>/<experiment-timestamp>`
    Returns:
    ----------
        params (dict): the loaded parameters
    '''
    with open(exp_param_filename(path)) as f:
        return yaml.safe_load(f)
    

def save_exp_params(path, params, logger=None):
    '''
    Arguments:
    ----------
        path (str): full path to experiment, i.e. 
            `<path-to-experiments-dir>/<experiment-timestamp>`
        params (dict): dictionary of parameters to save
        logger (logging.Logger): Defaults to None.
    '''
    with open(exp_param_filename(path), 'w') as f:
        yaml.dump(params, f, default_flow_style=False)
    if logger: 
        logger.info('Experiment parameters saved.')

def setup_exp(path, params, name=None):
    '''
    Arguments:
    ----------
        path (str): full path to where to create experiments, i.e. 
            `<path-to-experiments-dir>`
        params (dict): dictionary of parameters to save
    Returns:
    ----------
        exp_dir (str): full path to experiment, i.e. 
            `<path-to-experiments-dir>/<experiment-timestamp>`
        logger (logging.Logger)
    '''
    exp_name = gen_exp_name(name)
    exp_dir = os.path.join(path, exp_name)
    if not os.path.isdir(exp_dir):
        os.makedirs(exp_dir)

    logger = config_exp_logger(exp_dir)    
    save_exp_params(exp_dir, params, logger)
    return exp_dir, logger

    
def is_config_subset(truth, params):
    '''
    Arguments:
    ----------
        truth (dict): dictionary of parameters to compare to
        params (dict): dictionary of parameters to test
    Returns:
    ----------
        result (bool) whether or not `params` is a subset of `truth`
    '''
    if not type(truth) == type(params): return False
    for key, val in params.items():
        if key not in truth: return False
        if type(val) is dict:
            if not is_config_subset(truth[key], val): 
                return False
        else:            
            if not truth[key] == val: return False
    return True


def find_exps(path, params):
    '''
    Arguments:
    ----------
        path (str): full path to where to create experiments, i.e. 
            `<path-to-experiments-dir>`
        params (dict): dictionary of parameters to test
    Returns:
    ----------
        results (str[]) list of experiment names where their parameters are 
            supersets of the provided `params`
    '''
    exps = list_exps(path)
    results = []
    for exp in exps:
        exp_name = os.path.join(path, exp)
        exp_params = load_exp_params(exp_param_filename(exp_name))
        if is_config_subset(exp_params, params):
            results.append(exp)
    return results

# %% ../nbs/02_utils.ipynb 88
import os, pathlib, itertools
from pathlib import Path
from dataclasses import dataclass, field, KW_ONLY
from typing import Optional, List, ClassVar, Any, TypeAlias, Union

from ipos.imp import is_mod, is_var_imp, Module

# %% ../nbs/02_utils.ipynb 89
from iza.types import (
    PathLike, PathType,
    RichConsole, RichProgress, RichText, RichTree
)
from .static import EXT_PY
from .imp import RichImp

# %% ../nbs/02_utils.ipynb 93
@dataclass
class ArchiveDownloader:    
    _: KW_ONLY
    rootdir: str
    archive: str
    entries: Union[str, list[str]]
    savedir: str
    extract: bool = False
    cleanup: bool = False
    compound_archive: bool = False
    archives: Optional[list[str]] = None
    console: Optional[RichConsole] = None
    progress: Optional[RichProgress] = None

    _rich: Module = field(init=False, repr=False, default=None)
    has_rich: bool = field(init=False, repr=False, default=None)

    

    def __post_init__(self):    
        r = RichImp()
        self._rich = r._module
        self.has_rich = is_mod(self._rich)

        self.entries = self.entries if isinstance(self.entries, list) else [self.entries]        
        self.console = get_console()        
        self.progress = self.get_progress()
    
        self.savedir = Path(self.savedir).expanduser()
        make_missing_dirs(self.savedir)

    def get_progress(self):
        progress = getattr(self, 'progress', None)
        if progress is not None:
            return progress

        if self.has_rich and is_var_imp('Progress'):
                self.progress = Progress(console=self.console)
                return self.progress

        return None

    @property
    def path(self) -> str:
        return urljoin(self.rootdir, self.archive)

    @property
    def urls(self) -> list[str]:
        urls = []
        if self.compound_archive and self.archives is not None:
            for archive, entry in itertools.product(self.archives, self.entries):
                urls.append(urljoin(self.rootdir, archive, entry))
        else:
            urls = [urljoin(self.path, entry) for entry in self.entries]
        return urls

    def download_missing_files(self) -> None:        
        total_files = len(self.urls)
        if self.has_rich and self.progress:
            with self.progress:
                task = self.progress.add_task("[cyan]Downloading...", total=total_files)
                for url in self.urls:
                    filename = Path(parse_url(url).path).name
                    fullpath = self.savedir / filename
                    if not fullpath.exists():
                        stream_file(url, str(fullpath))
                        self.progress.advance(task)
                    else:
                        self.progress.advance(task)
        else:            
            for url in tqdm(self.urls, desc='Downloading'):       
                filename = Path(parse_url(url).path).name
                fullpath = self.savedir / filename
                if not fullpath.exists():
                    stream_file(url, str(fullpath))
                

    def calc_n_to_extract(self) -> int:
        files = [self.savedir / entry for entry in self.entries]
        files = get_gz_files_in_dir(self.savedir)
        total = 0
        for file in files:
            if is_tarball(file):
                total += 1
            elif is_gz(file):
                total += 1
        return total                

    def extract_files(self) -> None:
        recurser = RecursiveDecompressor(
            dirname=self.savedir, 
            entries=self.entries,
            strategy=undo_gz,
            remove=self.cleanup,             
            progress=self.progress
        )
        recurser.decompress()
        return
        files = [self.savedir / entry for entry in self.entries]
        files = get_gz_files_in_dir(self.savedir)
        total = self.calc_n_to_extract()
        if self.has_rich and self.progress:
            with self.progress:
                task = self.progress.add_task("[cyan]Extracting...", total=total)
                for file in files:
                    undo_gz(file, remove=self.cleanup)                    
                    self.progress.advance(task)
        else:
            for file in tqdm(files, desc='Extracting'):
                undo_gz(file, remove=self.cleanup)

    def execute(self) -> None:
        if self.has_rich and self.console:
            self.console.print(f"Processing archive: [bold cyan]{self.archive}[/bold cyan]")
        else:
            print(f"Processing archive: {self.archive}")

        self.download_missing_files()
        if self.extract:
            self.extract_files()

        if self.has_rich and self.console:
            dir = RichDirectory(self.savedir, console=self.console)
            dir.print_rich()
        else:
            dir = Directory(self.savedir)
            dir.print()

# %% ../nbs/02_utils.ipynb 97
#| eval: False
try:
    import typer
    app = typer.Typer()

    @app.command()
    def download(rootdir: str, archive: str, entries: List[str], savedir: str, extract: bool = False, cleanup: bool = False):
        downloader = ArchiveDownloader(rootdir, archive, entries, savedir, extract, cleanup)
        downloader.execute()

except ImportError:
    pass



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_imp.ipynb.

# %% auto 0
__all__ = ['TorchImp', 'TqdmImp', 'RichImp', 'ScanPyImp', 'SCPrepImp', 'is_rich_available', 'is_pytorch_available',
           'is_anndata_available']

# %% ../nbs/00_imp.ipynb 4
from ipos.imp import Imp, ImpSubSpec, ImpItem, ImpSubSpecType, is_mod_imp, is_mod, VariableDict
from dataclasses import dataclass, field

# %% ../nbs/00_imp.ipynb 6
@dataclass
class TorchImp(Imp):
    name: str = 'torch'
    lazy: bool = False
    subspecs: ImpSubSpecType = field(default_factory = lambda: [])

@dataclass
class TqdmImp(Imp):
    name: str = 'tqdm'
    subspecs: ImpSubSpecType = field(default_factory = lambda: [
        ImpSubSpec('tqdm', 'auto', [ImpItem('tqdm')])
    ])

@dataclass
class RichImp(Imp):
    name: str = 'rich'
    lazy: bool = False
    namespace: VariableDict = field(default_factory=globals, repr=False)
    subspecs: ImpSubSpecType = field(default_factory = lambda: [
        ImpSubSpec.from_str('from rich.tree import Tree'),
        ImpSubSpec.from_str('from rich.text import Text'),
        ImpSubSpec.from_str('from rich.markup import espace'),
        ImpSubSpec.from_str('from rich.filesize import decimal'),
        ImpSubSpec.from_str('from rich.filesize import Console'),
        ImpSubSpec.from_str('from rich.progress import Progress'),    
        ImpSubSpec('rich', '', [ImpItem('get_console')])
    ])

@dataclass
class ScanPyImp(Imp):
    name: str = 'scanpy'
    nick: str = 'sc'
    subspecs: ImpSubSpecType = field(default_factory = lambda: [
        ImpSubSpec('scanpy', '', [ImpItem('AnnData'), ImpItem('read_h5ad')]),
    ])

@dataclass
class SCPrepImp(Imp):
    name: str = 'scprep'    
    subspecs: ImpSubSpecType = field(default_factory = lambda: [
        ImpSubSpec('scanpy', 'io', [ImpItem('load_mtx')]),
    ])

# %% ../nbs/00_imp.ipynb 8
def is_rich_available() -> bool:
    try:
        import rich
        return True
    except ImportError:
        return False

def is_pytorch_available() -> bool: 
    try:
        import torch
        return True
    except ImportError:
        return False

def is_anndata_available() -> bool:    
    try:
        import anndata 
        return True
    except ImportError:
        return False



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_static.ipynb.

# %% auto 0
__all__ = ['NBKS_DIR', 'ROOT_DIR', 'EXPL_DIR', 'DATA_DIR', 'RESULTS_DIR', 'FIGURES_DIR', 'IZALITH_DATA_DIR', 'GIBBS_DATA_DIR',
           'PERIOD', 'GZ', 'H5', 'PT', 'PY', 'BED', 'CSV', 'MTX', 'TAR', 'TSV', 'NPY', 'NPZ', 'TXT', 'PIP', 'PTH',
           'ZIP', 'JSON', 'H5AD', 'IPYNB', 'ARROW', 'FASTA', 'PARQUET', 'EXT_GZ', 'EXT_H5', 'EXT_PT', 'EXT_PY',
           'EXT_BED', 'EXT_CSV', 'EXT_MTX', 'EXT_TAR', 'EXT_TSV', 'EXT_NPY', 'EXT_NPZ', 'EXT_TXT', 'EXT_PIP', 'EXT_PTH',
           'EXT_ZIP', 'EXT_JSON', 'EXT_H5AD', 'EXT_IPYNB', 'EXT_ARROW', 'EXT_FASTA', 'EXT_PARQUET', 'EXT_TAR_GZ',
           'SEED', 'ITEM', 'TIME', 'LABEL', 'SERIES', 'SAMPLE', 'SAMPLES', 'CONDITION', 'CONDITIONS', 'OBS', 'VAR',
           'OBSM', 'VARM', 'OBSP', 'VARP', 'UNS', 'LAYER', 'LAYERS', 'HVG', 'PCA', 'UMAP', 'TSNE', 'PHATE', 'MAGIC',
           'CELL', 'CELLS', 'BARCODE', 'BARCODES', 'ID', 'MITO', 'RIBO', 'GENE', 'GENE_ID', 'GENE_IDS', 'GENE_SYMBOL',
           'HIGHLY_VARIABLE', 'ENSEMBL', 'ENSEMBL_ID', 'HUMAN', 'MOUSE', 'HUMAN_TF', 'MOUSE_TF', 'HUMAN_GENE_SYMBOL',
           'HUMAN_ENSEMBLE_ID', 'MOUSE_ENSEMBLE_ID', 'BATCH', 'TIMEPOINT', 'TIMEPOINTS', 'TOTAL_COUNTS', 'DOUBLET',
           'DOUBLET_SCORES', 'PREDICTED_DOUBLETS', 'PRENORM', 'DETECTED', 'SCALED_NORMALIZED', 'COUNTS', 'PCT',
           'PERCENT', 'PCT_COUNTS', 'PCT_COUNTS_MITO', 'PCT_COUNTS_RIBO', 'X', 'X_', 'X_PRENORM', 'X_DETECTED',
           'X_SCALED_NORMALIZED', 'X_MAGIC', 'X_PCA', 'X_PCA_HVG', 'X_PHATE', 'X_PHATE_HVG', 'ADATA', 'MATRIX',
           'FEATURES', 'SENSITIVITY', 'AMAZON_BUCKET_FLUENTBIO', 'LINEAGE_ESC', 'LINEAGE_PREGERM',
           'LINEAGE_NEURAL_CREST', 'LINEAGE_NEURAL_PROG', 'LINEAGE_ENDO', 'LINEAGE_MESO', 'LINEAGE_CARDIAC',
           'LINEAGE_HEMA', 'LINEAGE_SOMITES', 'LINEAGE_TROPHOBLAST', 'LINEAGE_ALL', 'LINEAGE_NAMES', 'MARKERS_ESC',
           'MARKERS_PREGERM', 'MARKERS_NEURAL_CREST', 'MARKERS_NEURAL_PROG', 'MARKERS_ENDO', 'MARKERS_MESO',
           'MARKERS_CARDIAC', 'MARKERS_HEMA', 'MARKERS_SOMITES', 'MARKERS_TROPHOBLAST', 'MARKERS_LISTS', 'MARKERS_ALL',
           'LINEAGE_MARKERS', 'RC_PARAMS', 'SEQRUN', 'SEQRUNS', 'TREATMENT', 'TREATMENTS', 'SENS', 'SENSITIVITIES']

# %% ../nbs/01_static.ipynb 5
import os, sys, pkg_resources

# %% ../nbs/01_static.ipynb 6
'''----------------------------------------------------
NOTE: RELATIVE DEFINITIONS
----------------------------------------------------'''

# nbdev nbs directory, i.e. `nbs/``
NBKS_DIR = os.path.abspath('')

# project root directory
ROOT_DIR = os.path.join(NBKS_DIR, '..')

# where example notebooks are stored, i.e. `examples/`
EXPL_DIR = os.path.join(ROOT_DIR, 'examples')

# where small data is stored, i.e. `data/`
DATA_DIR = os.path.join(ROOT_DIR, 'data')

# where small data is stored, i.e. `results/`
RESULTS_DIR = os.path.join(ROOT_DIR, 'results')

# where small data is stored, i.e. `figures/`
FIGURES_DIR = os.path.join(ROOT_DIR, 'figures')

IZALITH_DATA_DIR = os.path.expanduser('~/Data')

GIBBS_DATA_DIR = '/gpfs/gibbs/pi/krishnaswamy_smita'

# %% ../nbs/01_static.ipynb 9
PERIOD = '.'

# %% ../nbs/01_static.ipynb 10
GZ = f'gz'
H5 = f'h5'
PT = f'pt'
PY = f'py'

BED = f'bed'
CSV = f'csv'
MTX = f'mtx'
TAR = f'tar'
TSV = f'tsv'
NPY = f'npy'
NPZ = f'npz'
TXT = f'txt'
PIP = f'pip'
PTH = f'pth'
ZIP = f'zip'

JSON = f'json'
H5AD = f'h5ad'
IPYNB = f'ipynb'

ARROW = 'arrow'
FASTA = 'fasta'
PARQUET = 'parquet'


# %% ../nbs/01_static.ipynb 11
EXT_GZ = f'{PERIOD}{GZ}'
EXT_H5 = f'{PERIOD}{H5}'
EXT_PT = f'{PERIOD}{PT}'
EXT_PY = f'{PERIOD}{PY}'

EXT_BED = f'{PERIOD}{BED}'
EXT_CSV = f'{PERIOD}{CSV}'
EXT_MTX = f'{PERIOD}{MTX}'
EXT_TAR = f'{PERIOD}{TAR}'
EXT_TSV = f'{PERIOD}{TSV}'
EXT_NPY = f'{PERIOD}{NPY}'
EXT_NPZ = f'{PERIOD}{NPZ}'
EXT_TXT = f'{PERIOD}{TXT}'
EXT_PIP = f'{PERIOD}{PIP}'
EXT_PTH = f'{PERIOD}{PTH}'
EXT_ZIP = f'{PERIOD}{ZIP}'

EXT_JSON = f'{PERIOD}{JSON}'
EXT_H5AD = f'{PERIOD}{H5AD}'
EXT_IPYNB = f'{PERIOD}{IPYNB}'

EXT_ARROW = f'{PERIOD}{ARROW}'
EXT_FASTA = f'{PERIOD}{FASTA}'

EXT_PARQUET = f'{PERIOD}{PARQUET}'

EXT_TAR_GZ = f'{EXT_TAR}{EXT_GZ}'

# %% ../nbs/01_static.ipynb 15
SEED = 'seed'
ITEM = 'item'
TIME = 'time'
LABEL = 'label'
SERIES = 'series'
SAMPLE = 'sample'
SAMPLES = 'samples'
CONDITION = 'condition'
CONDITIONS = f'{CONDITION}s'

# %% ../nbs/01_static.ipynb 19
OBS = 'obs'
VAR = 'var'

OBSM = f'{OBS}m'
VARM = f'{VAR}m'

OBSP = f'{OBS}p'
VARP = f'{VAR}p'

UNS = 'UNS'
LAYER = 'layer'
LAYERS = 'layers'

# %% ../nbs/01_static.ipynb 21
HVG = 'hvg'

PCA = 'pca'
UMAP = 'umap'
TSNE = 'tsne'
PHATE = 'phate'
MAGIC = 'magic'

# %% ../nbs/01_static.ipynb 23
CELL = 'cell'
CELLS = f'{CELL}s'
BARCODE = 'barcode'
BARCODES = f'{BARCODE}s'

# %% ../nbs/01_static.ipynb 25
ID = 'id'
MITO = 'mito'
RIBO = 'ribo'
GENE = 'gene'
GENE_ID = f'{GENE}_id'
GENE_IDS = f'{GENE_ID}s'
GENE_SYMBOL = f'{GENE}_symbol'
HIGHLY_VARIABLE = 'highly_variable'

ENSEMBL = 'ensembl'
ENSEMBL_ID = f'{ENSEMBL}_id'

HUMAN = 'human'
MOUSE = 'mouse'

HUMAN_TF = f'{HUMAN}TFs'
MOUSE_TF = f'{MOUSE}TFs'

HUMAN_GENE_SYMBOL = f'{HUMAN}GeneSymbol'
HUMAN_ENSEMBLE_ID = f'{HUMAN}EnsemblID'
MOUSE_ENSEMBLE_ID = f'{MOUSE}EnsemblID'

BATCH = 'batch'
TIMEPOINT = 'timepoint'
TIMEPOINTS = f'{TIMEPOINT}s'
TOTAL_COUNTS = 'total_counts'
DOUBLET = 'doublet'
DOUBLET_SCORES = f'{DOUBLET}_scores'
PREDICTED_DOUBLETS = f'predicted_{DOUBLET}s'

PRENORM = 'prenorm'
DETECTED = 'detected'
SCALED_NORMALIZED = 'scaled_normalized'


COUNTS = 'counts'
PCT = 'pct'
PERCENT = 'percent'
PCT_COUNTS = f'{PCT}_{COUNTS}'
PCT_COUNTS_MITO = f'{PCT}_{COUNTS}_{MITO}'
PCT_COUNTS_RIBO = f'{PCT}_{COUNTS}_{RIBO}'

# %% ../nbs/01_static.ipynb 26
_ = '_'
X = 'X'
X_ = f'{X}{_}'

# %% ../nbs/01_static.ipynb 27
# Layer storing prenormalized counts in adata.layers
X_PRENORM = f'{X_}{PRENORM}'

# Layer storing values where prenormalized counts is at least 0 in adata.layers
X_DETECTED = f'{X_}{DETECTED}'

# Layer storing counts normalized
X_SCALED_NORMALIZED = f'{X_}{SCALED_NORMALIZED}'

# Layer storing MAGIC embedding in adata.obsm
X_MAGIC = f'{X_}{MAGIC}'

# Layer storing PCA embedding in adata.obsm
X_PCA = f'{X_}{PCA}'

# Layer storing PCA embedding just on HVGs in adata.obsm
X_PCA_HVG = f'{X_}{PCA}_{HVG}'

# Layer storing PHATE embedding just in adata.obsm
X_PHATE = f'{X_}{PHATE}'

# Layer storing PHATE embedding just on HVGs in adata.obsm
X_PHATE_HVG = f'{X_}{PHATE}_{HVG}'

# %% ../nbs/01_static.ipynb 29
ADATA = 'adata'
MATRIX = 'matrix'
FEATURES = 'features'
SENSITIVITY = 'sensitivity'

# %% ../nbs/01_static.ipynb 31
AMAZON_BUCKET_FLUENTBIO = 'https://fbs-public.s3.us-east-2.amazonaws.com'

# %% ../nbs/01_static.ipynb 33
import itertools

# %% ../nbs/01_static.ipynb 35
LINEAGE_ESC = 'esc'
LINEAGE_PREGERM = 'pregerm'
LINEAGE_NEURAL_CREST = 'neural_crest'
LINEAGE_NEURAL_PROG = 'neural_prog'
LINEAGE_ENDO = 'endo'
LINEAGE_MESO = 'meso'
LINEAGE_CARDIAC = 'cardiac'
LINEAGE_HEMA = 'hema'
LINEAGE_SOMITES = 'somites'
LINEAGE_TROPHOBLAST = 'trophblast'
LINEAGE_ALL = 'all'

# %% ../nbs/01_static.ipynb 36
LINEAGE_NAMES = [
    LINEAGE_ESC, LINEAGE_PREGERM, LINEAGE_NEURAL_CREST,
    LINEAGE_NEURAL_PROG, LINEAGE_ENDO, LINEAGE_MESO,
    LINEAGE_CARDIAC, LINEAGE_HEMA, LINEAGE_SOMITES,
    LINEAGE_TROPHOBLAST, LINEAGE_ALL
]

# %% ../nbs/01_static.ipynb 38
MARKERS_ESC = ['POU5F1', 'PRDM14', 'NANOG', 'SOX2', 'DPPA4', 'NR2F6', 'OTX2']
MARKERS_PREGERM = ['TP53','NR6A1','GBX2','ZIC2','ZIC5']
MARKERS_NEURAL_CREST = ['PAX3','FOXD3','ITGA4','EDNRB','SOX9','SOX10']
MARKERS_NEURAL_PROG = ['NES','MAP2','PAX6','SOX1','ZBTB16','ONECUT1', 'ONECUT2','NR2F1','ISL1','NPAS1','EN2','SOX2','NEUROD1']
MARKERS_ENDO = ['EOMES','FOXA2','SOX17','GATA4','GATA6','HHEX']
MARKERS_MESO = ['MIXL1','CER1']
MARKERS_CARDIAC = ['NKX2-5','GATA4','MYBPC2','TTN','TNNT2','MYH6','TBX5']
MARKERS_HEMA = ['NKX2-5','TAL1','SOX17','CD34','PECAM1','CD69','CXCR4','CDH5','KDR','LMO2']
MARKERS_SOMITES = ['WT1','TBX18','TBX15','PDGFRA','SIX2','TBX5']
MARKERS_TROPHOBLAST = ['GATA3','TP63','ITGA6','CDH1']

# %% ../nbs/01_static.ipynb 39
MARKERS_LISTS =  [
    MARKERS_ESC, MARKERS_PREGERM, MARKERS_NEURAL_CREST,
    MARKERS_NEURAL_PROG, MARKERS_ENDO, MARKERS_MESO, 
    MARKERS_CARDIAC, MARKERS_HEMA, MARKERS_SOMITES, 
    MARKERS_TROPHOBLAST,
]
MARKERS_ALL = list(itertools.chain(*MARKERS_LISTS))
MARKERS_LISTS.append(MARKERS_ALL)

# %% ../nbs/01_static.ipynb 41
LINEAGE_MARKERS = dict(zip(LINEAGE_NAMES, MARKERS_LISTS))

# %% ../nbs/01_static.ipynb 44
RC_PARAMS = {
    'axes.axisbelow': False,
    'axes.edgecolor': 'lightgrey',
    'axes.facecolor': 'None',
    'axes.grid': False,
    'axes.labelcolor': 'dimgrey',
    'axes.spines.right': False,
    'axes.spines.top': False,
    'figure.facecolor': 'white',
    'lines.solid_capstyle': 'round',
    'patch.edgecolor': 'w',
    'patch.force_edgecolor': True,
    'text.color': 'dimgrey',
    'xtick.bottom': False,
    'xtick.color': 'dimgrey',
    'xtick.direction': 'out',
    'xtick.top': False,
    'ytick.color': 'dimgrey',
    'ytick.direction': 'out',
    'ytick.left': False,
    'ytick.right': False,
    'font.size': 12,
    'axes.titlesize': 10,
    'axes.labelsize': 12
}

# %% ../nbs/01_static.ipynb 48
SEQRUN = 'seqrun'
SEQRUNS = f'{SEQRUN}s'
TREATMENT = f'treatment'
TREATMENTS = f'{TREATMENT}s'

SENS = 'sens'
SENSITIVITIES = 'sensitivities'



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/99_commands.ipynb.

# %% auto 0
__all__ = ['nbdev_agg_app', 'aggregate_notebooks', 'callback']

# %% ../nbs/99_commands.ipynb 3
import os, typer
from rich.pretty import pprint
from typing import Optional, List, Union
from typing_extensions import Annotated

# %% ../nbs/99_commands.ipynb 4
from .rich import console
from .typer import app
from .nbs import NotebookAggregator

# %% ../nbs/99_commands.ipynb 5
nbdev_agg_app = typer.Typer()

@app.command('nbdev_agg')
def aggregate_notebooks(
    path: Annotated[str, typer.Option(
        '--path', '-p', help='The path to the directory of notebooks to aggregate.',
    )],
    module: Annotated[Optional[str], typer.Option(
        '--module', '-m', help='The name of the module to create.'
    )] =  None,
    output: Annotated[Optional[str], typer.Option(
        '--output', '-o', help='The path to where the save the aggregated notebook file name excluded.',
    )] =  None,    
    ignore: Annotated[Optional[List[str]], typer.Option(
        '--ignore', '-i', help='Name of files to ignore in the directory.',
    )] =  [],
    prefix: Annotated[Optional[bool], typer.Option(
        '--prefix', '-f', help='Whether to prefix the module name to the notebook names.',
    )] =  True, 
):
    if ignore is None:
        ignore = []

    aggregator = NotebookAggregator(path, module=module, output=output, ignore=ignore, prefix=prefix)
    aggregator.aggregate()
    console.print(f"Aggregated notebook saved to {aggregator.output}", style="bold green")


@app.callback()
def callback():
    """
    Aggregate the notebooks in the directory.

    Parameters
    ----------
    path : str
        The path to the directory containing the notebooks.
    module : Optional[str], optional
        The name of the module, by default None
    output : Optional[str], optional
        The path to the output notebook, by default None
    ignore : Optional[List[str]], optional
        A list of notebooks to ignore, by default None
    """
    pass




# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/iza',
                'doc_host': 'https://dsm-72.github.io',
                'git_url': 'https://github.com/dsm-72/iza',
                'lib_path': 'iza'},
  'syms': { 'iza.commands': { 'iza.commands.aggregate_notebooks': ('commands.html#aggregate_notebooks', 'iza/commands.py'),
                              'iza.commands.callback': ('commands.html#callback', 'iza/commands.py')},
            'iza.imp': { 'iza.imp.RichImp': ('imp.html#richimp', 'iza/imp.py'),
                         'iza.imp.SCPrepImp': ('imp.html#scprepimp', 'iza/imp.py'),
                         'iza.imp.ScanPyImp': ('imp.html#scanpyimp', 'iza/imp.py'),
                         'iza.imp.TorchImp': ('imp.html#torchimp', 'iza/imp.py'),
                         'iza.imp.TqdmImp': ('imp.html#tqdmimp', 'iza/imp.py'),
                         'iza.imp.is_anndata_available': ('imp.html#is_anndata_available', 'iza/imp.py'),
                         'iza.imp.is_pytorch_available': ('imp.html#is_pytorch_available', 'iza/imp.py'),
                         'iza.imp.is_rich_available': ('imp.html#is_rich_available', 'iza/imp.py')},
            'iza.nbs': { 'iza.nbs.NotebookAggregator': ('nbs.html#notebookaggregator', 'iza/nbs.py'),
                         'iza.nbs.NotebookAggregator.__post_init__': ('nbs.html#notebookaggregator.__post_init__', 'iza/nbs.py'),
                         'iza.nbs.NotebookAggregator.add_current_notebook_section': ( 'nbs.html#notebookaggregator.add_current_notebook_section',
                                                                                      'iza/nbs.py'),
                         'iza.nbs.NotebookAggregator.aggregate': ('nbs.html#notebookaggregator.aggregate', 'iza/nbs.py'),
                         'iza.nbs.NotebookAggregator.basename': ('nbs.html#notebookaggregator.basename', 'iza/nbs.py'),
                         'iza.nbs.NotebookAggregator.dirname': ('nbs.html#notebookaggregator.dirname', 'iza/nbs.py'),
                         'iza.nbs.NotebookAggregator.format_header': ('nbs.html#notebookaggregator.format_header', 'iza/nbs.py'),
                         'iza.nbs.NotebookAggregator.get_prefix': ('nbs.html#notebookaggregator.get_prefix', 'iza/nbs.py'),
                         'iza.nbs.NotebookAggregator.nb_as_href': ('nbs.html#notebookaggregator.nb_as_href', 'iza/nbs.py'),
                         'iza.nbs.NotebookAggregator.notebooks': ('nbs.html#notebookaggregator.notebooks', 'iza/nbs.py'),
                         'iza.nbs.NotebookAggregator.write_nbdev_head': ('nbs.html#notebookaggregator.write_nbdev_head', 'iza/nbs.py'),
                         'iza.nbs.NotebookAggregator.write_nbdev_tail': ('nbs.html#notebookaggregator.write_nbdev_tail', 'iza/nbs.py'),
                         'iza.nbs.NotebookFinder': ('nbs.html#notebookfinder', 'iza/nbs.py'),
                         'iza.nbs.NotebookFinder.__init__': ('nbs.html#notebookfinder.__init__', 'iza/nbs.py'),
                         'iza.nbs.NotebookFinder.find_module': ('nbs.html#notebookfinder.find_module', 'iza/nbs.py'),
                         'iza.nbs.NotebookLoader': ('nbs.html#notebookloader', 'iza/nbs.py'),
                         'iza.nbs.NotebookLoader.__init__': ('nbs.html#notebookloader.__init__', 'iza/nbs.py'),
                         'iza.nbs.NotebookLoader.load_module': ('nbs.html#notebookloader.load_module', 'iza/nbs.py'),
                         'iza.nbs.NotebookLogger': ('nbs.html#notebooklogger', 'iza/nbs.py'),
                         'iza.nbs.NotebookLogger.__del__': ('nbs.html#notebooklogger.__del__', 'iza/nbs.py'),
                         'iza.nbs.NotebookLogger.__post_init__': ('nbs.html#notebooklogger.__post_init__', 'iza/nbs.py'),
                         'iza.nbs.NotebookPrefixer': ('nbs.html#notebookprefixer', 'iza/nbs.py'),
                         'iza.nbs.NotebookPrefixer.__post_init__': ('nbs.html#notebookprefixer.__post_init__', 'iza/nbs.py'),
                         'iza.nbs.NotebookPrefixer.filename_without_prefix': ( 'nbs.html#notebookprefixer.filename_without_prefix',
                                                                               'iza/nbs.py'),
                         'iza.nbs.NotebookPrefixer.int_to_prefix_int': ('nbs.html#notebookprefixer.int_to_prefix_int', 'iza/nbs.py'),
                         'iza.nbs.NotebookPrefixer.next_nb_prefix': ('nbs.html#notebookprefixer.next_nb_prefix', 'iza/nbs.py'),
                         'iza.nbs.NotebookPrefixer.prefix_from_filename': ('nbs.html#notebookprefixer.prefix_from_filename', 'iza/nbs.py'),
                         'iza.nbs.NotebookPrefixer.prefix_int_to_int': ('nbs.html#notebookprefixer.prefix_int_to_int', 'iza/nbs.py'),
                         'iza.nbs.NotebookViewer': ('nbs.html#notebookviewer', 'iza/nbs.py'),
                         'iza.nbs.NotebookViewer.__init__': ('nbs.html#notebookviewer.__init__', 'iza/nbs.py'),
                         'iza.nbs.NotebookViewer.__post_init__': ('nbs.html#notebookviewer.__post_init__', 'iza/nbs.py'),
                         'iza.nbs.NotebookViewer.get_cell': ('nbs.html#notebookviewer.get_cell', 'iza/nbs.py'),
                         'iza.nbs.NotebookViewer.prepare_cell': ('nbs.html#notebookviewer.prepare_cell', 'iza/nbs.py'),
                         'iza.nbs.NotebookViewer.show_cell': ('nbs.html#notebookviewer.show_cell', 'iza/nbs.py'),
                         'iza.nbs.NotebookViewer.show_notebook': ('nbs.html#notebookviewer.show_notebook', 'iza/nbs.py'),
                         'iza.nbs.find_notebook': ('nbs.html#find_notebook', 'iza/nbs.py')},
            'iza.plots': { 'iza.plots.convert_cmap_to_categorical': ('plots.html#convert_cmap_to_categorical', 'iza/plots.py'),
                           'iza.plots.hide_axes': ('plots.html#hide_axes', 'iza/plots.py'),
                           'iza.plots.make_categorical_legend': ('plots.html#make_categorical_legend', 'iza/plots.py'),
                           'iza.plots.mpl_cat_val': ('plots.html#mpl_cat_val', 'iza/plots.py'),
                           'iza.plots.mpl_cur_cmap': ('plots.html#mpl_cur_cmap', 'iza/plots.py'),
                           'iza.plots.scale_figsize_for_grid': ('plots.html#scale_figsize_for_grid', 'iza/plots.py'),
                           'iza.plots.set_rc_params': ('plots.html#set_rc_params', 'iza/plots.py'),
                           'iza.plots.setup_grid_figure': ('plots.html#setup_grid_figure', 'iza/plots.py'),
                           'iza.plots.triple_scatter': ('plots.html#triple_scatter', 'iza/plots.py')},
            'iza.rich': {'iza.rich.cprint': ('rich.html#cprint', 'iza/rich.py'), 'iza.rich.swrap': ('rich.html#swrap', 'iza/rich.py')},
            'iza.static': {},
            'iza.typer': {},
            'iza.types': { 'iza.types.DirectoryTreeStrings': ('types.html#directorytreestrings', 'iza/types.py'),
                           'iza.types.TrainTestSplit': ('types.html#traintestsplit', 'iza/types.py'),
                           'iza.types.TrainValidSplit': ('types.html#trainvalidsplit', 'iza/types.py'),
                           'iza.types.TrainValidTestSplit': ('types.html#trainvalidtestsplit', 'iza/types.py')},
            'iza.utils': { 'iza.utils.AdataExtractor': ('utils.html#adataextractor', 'iza/utils.py'),
                           'iza.utils.AdataExtractor.axis_str': ('utils.html#adataextractor.axis_str', 'iza/utils.py'),
                           'iza.utils.AdataExtractor.df_cnt': ('utils.html#adataextractor.df_cnt', 'iza/utils.py'),
                           'iza.utils.AdataExtractor.df_emb': ('utils.html#adataextractor.df_emb', 'iza/utils.py'),
                           'iza.utils.AdataExtractor.emb_cols': ('utils.html#adataextractor.emb_cols', 'iza/utils.py'),
                           'iza.utils.AdataExtractor.get_df_cnt': ('utils.html#adataextractor.get_df_cnt', 'iza/utils.py'),
                           'iza.utils.AdataExtractor.get_df_emb': ('utils.html#adataextractor.get_df_emb', 'iza/utils.py'),
                           'iza.utils.AdataExtractor.get_emb': ('utils.html#adataextractor.get_emb', 'iza/utils.py'),
                           'iza.utils.AdataExtractor.get_layer': ('utils.html#adataextractor.get_layer', 'iza/utils.py'),
                           'iza.utils.AdataExtractor.has_emb': ('utils.html#adataextractor.has_emb', 'iza/utils.py'),
                           'iza.utils.AdataExtractor.has_hvg': ('utils.html#adataextractor.has_hvg', 'iza/utils.py'),
                           'iza.utils.AdataExtractor.sdata': ('utils.html#adataextractor.sdata', 'iza/utils.py'),
                           'iza.utils.ArchiveDownloader': ('utils.html#archivedownloader', 'iza/utils.py'),
                           'iza.utils.ArchiveDownloader.__post_init__': ('utils.html#archivedownloader.__post_init__', 'iza/utils.py'),
                           'iza.utils.ArchiveDownloader.calc_n_to_extract': ( 'utils.html#archivedownloader.calc_n_to_extract',
                                                                              'iza/utils.py'),
                           'iza.utils.ArchiveDownloader.download_missing_files': ( 'utils.html#archivedownloader.download_missing_files',
                                                                                   'iza/utils.py'),
                           'iza.utils.ArchiveDownloader.execute': ('utils.html#archivedownloader.execute', 'iza/utils.py'),
                           'iza.utils.ArchiveDownloader.extract_files': ('utils.html#archivedownloader.extract_files', 'iza/utils.py'),
                           'iza.utils.ArchiveDownloader.get_progress': ('utils.html#archivedownloader.get_progress', 'iza/utils.py'),
                           'iza.utils.ArchiveDownloader.path': ('utils.html#archivedownloader.path', 'iza/utils.py'),
                           'iza.utils.ArchiveDownloader.urls': ('utils.html#archivedownloader.urls', 'iza/utils.py'),
                           'iza.utils.Directory': ('utils.html#directory', 'iza/utils.py'),
                           'iza.utils.DirectoryTree': ('utils.html#directorytree', 'iza/utils.py'),
                           'iza.utils.DirectoryTree.__post_init__': ('utils.html#directorytree.__post_init__', 'iza/utils.py'),
                           'iza.utils.DirectoryTree.__repr__': ('utils.html#directorytree.__repr__', 'iza/utils.py'),
                           'iza.utils.DirectoryTree.get_tree_lines': ('utils.html#directorytree.get_tree_lines', 'iza/utils.py'),
                           'iza.utils.DirectoryTree.make_tree_str': ('utils.html#directorytree.make_tree_str', 'iza/utils.py'),
                           'iza.utils.DirectoryTree.print': ('utils.html#directorytree.print', 'iza/utils.py'),
                           'iza.utils.DirectoryTree.tree_generator': ('utils.html#directorytree.tree_generator', 'iza/utils.py'),
                           'iza.utils.RecursiveDecompressor': ('utils.html#recursivedecompressor', 'iza/utils.py'),
                           'iza.utils.RecursiveDecompressor.__post_init__': ( 'utils.html#recursivedecompressor.__post_init__',
                                                                              'iza/utils.py'),
                           'iza.utils.RecursiveDecompressor.decompress': ('utils.html#recursivedecompressor.decompress', 'iza/utils.py'),
                           'iza.utils.RecursiveDecompressor.decrease_total': ( 'utils.html#recursivedecompressor.decrease_total',
                                                                               'iza/utils.py'),
                           'iza.utils.RecursiveDecompressor.execute': ('utils.html#recursivedecompressor.execute', 'iza/utils.py'),
                           'iza.utils.RecursiveDecompressor.recurse': ('utils.html#recursivedecompressor.recurse', 'iza/utils.py'),
                           'iza.utils.RecursiveDecompressor.update_total': ( 'utils.html#recursivedecompressor.update_total',
                                                                             'iza/utils.py'),
                           'iza.utils.RichDirectory': ('utils.html#richdirectory', 'iza/utils.py'),
                           'iza.utils.RichDirectory.__post_init__': ('utils.html#richdirectory.__post_init__', 'iza/utils.py'),
                           'iza.utils.RichDirectory.print_rich': ('utils.html#richdirectory.print_rich', 'iza/utils.py'),
                           'iza.utils.Slice': ('utils.html#slice', 'iza/utils.py'),
                           'iza.utils.Slice.__post_init__': ('utils.html#slice.__post_init__', 'iza/utils.py'),
                           'iza.utils.Slice.astype': ('utils.html#slice.astype', 'iza/utils.py'),
                           'iza.utils.Slice.start': ('utils.html#slice.start', 'iza/utils.py'),
                           'iza.utils.Slice.step': ('utils.html#slice.step', 'iza/utils.py'),
                           'iza.utils.Slice.stop': ('utils.html#slice.stop', 'iza/utils.py'),
                           'iza.utils.Slice.todict': ('utils.html#slice.todict', 'iza/utils.py'),
                           'iza.utils.Slice.tolist': ('utils.html#slice.tolist', 'iza/utils.py'),
                           'iza.utils.Slice.toslice': ('utils.html#slice.toslice', 'iza/utils.py'),
                           'iza.utils.Slice.totuple': ('utils.html#slice.totuple', 'iza/utils.py'),
                           'iza.utils.allinstance': ('utils.html#allinstance', 'iza/utils.py'),
                           'iza.utils.allsametype': ('utils.html#allsametype', 'iza/utils.py'),
                           'iza.utils.arein': ('utils.html#arein', 'iza/utils.py'),
                           'iza.utils.base_entry_fn': ('utils.html#base_entry_fn', 'iza/utils.py'),
                           'iza.utils.base_init_tree': ('utils.html#base_init_tree', 'iza/utils.py'),
                           'iza.utils.check_ext': ('utils.html#check_ext', 'iza/utils.py'),
                           'iza.utils.collapse_user': ('utils.html#collapse_user', 'iza/utils.py'),
                           'iza.utils.config_exp_logger': ('utils.html#config_exp_logger', 'iza/utils.py'),
                           'iza.utils.copy_to_clipboard': ('utils.html#copy_to_clipboard', 'iza/utils.py'),
                           'iza.utils.decompress_directory_of_gunzipped_files': ( 'utils.html#decompress_directory_of_gunzipped_files',
                                                                                  'iza/utils.py'),
                           'iza.utils.decompress_gunzip': ('utils.html#decompress_gunzip', 'iza/utils.py'),
                           'iza.utils.decompress_tarball': ('utils.html#decompress_tarball', 'iza/utils.py'),
                           'iza.utils.decompress_tarball_of_gunzipped_files': ( 'utils.html#decompress_tarball_of_gunzipped_files',
                                                                                'iza/utils.py'),
                           'iza.utils.dir_dirs': ('utils.html#dir_dirs', 'iza/utils.py'),
                           'iza.utils.download_and_decompress_tarball_of_gunzipped_files': ( 'utils.html#download_and_decompress_tarball_of_gunzipped_files',
                                                                                             'iza/utils.py'),
                           'iza.utils.drop_ext': ('utils.html#drop_ext', 'iza/utils.py'),
                           'iza.utils.entry_fn': ('utils.html#entry_fn', 'iza/utils.py'),
                           'iza.utils.exp_log_filename': ('utils.html#exp_log_filename', 'iza/utils.py'),
                           'iza.utils.exp_param_filename': ('utils.html#exp_param_filename', 'iza/utils.py'),
                           'iza.utils.filter_for_gz_files': ('utils.html#filter_for_gz_files', 'iza/utils.py'),
                           'iza.utils.filter_kwargs_for_class': ('utils.html#filter_kwargs_for_class', 'iza/utils.py'),
                           'iza.utils.filter_kwargs_for_func': ('utils.html#filter_kwargs_for_func', 'iza/utils.py'),
                           'iza.utils.find_exps': ('utils.html#find_exps', 'iza/utils.py'),
                           'iza.utils.gen_exp_name': ('utils.html#gen_exp_name', 'iza/utils.py'),
                           'iza.utils.get_func_params': ('utils.html#get_func_params', 'iza/utils.py'),
                           'iza.utils.get_gz_files_in_dir': ('utils.html#get_gz_files_in_dir', 'iza/utils.py'),
                           'iza.utils.get_user': ('utils.html#get_user', 'iza/utils.py'),
                           'iza.utils.init_tree': ('utils.html#init_tree', 'iza/utils.py'),
                           'iza.utils.is_config_subset': ('utils.html#is_config_subset', 'iza/utils.py'),
                           'iza.utils.is_cpu': ('utils.html#is_cpu', 'iza/utils.py'),
                           'iza.utils.is_device': ('utils.html#is_device', 'iza/utils.py'),
                           'iza.utils.is_gz': ('utils.html#is_gz', 'iza/utils.py'),
                           'iza.utils.is_matrix': ('utils.html#is_matrix', 'iza/utils.py'),
                           'iza.utils.is_mps': ('utils.html#is_mps', 'iza/utils.py'),
                           'iza.utils.is_np': ('utils.html#is_np', 'iza/utils.py'),
                           'iza.utils.is_path': ('utils.html#is_path', 'iza/utils.py'),
                           'iza.utils.is_pathlib': ('utils.html#is_pathlib', 'iza/utils.py'),
                           'iza.utils.is_pathlike': ('utils.html#is_pathlike', 'iza/utils.py'),
                           'iza.utils.is_series': ('utils.html#is_series', 'iza/utils.py'),
                           'iza.utils.is_series_like': ('utils.html#is_series_like', 'iza/utils.py'),
                           'iza.utils.is_tar': ('utils.html#is_tar', 'iza/utils.py'),
                           'iza.utils.is_tarball': ('utils.html#is_tarball', 'iza/utils.py'),
                           'iza.utils.is_targz': ('utils.html#is_targz', 'iza/utils.py'),
                           'iza.utils.is_tensor': ('utils.html#is_tensor', 'iza/utils.py'),
                           'iza.utils.is_torch': ('utils.html#is_torch', 'iza/utils.py'),
                           'iza.utils.isin': ('utils.html#isin', 'iza/utils.py'),
                           'iza.utils.isiter': ('utils.html#isiter', 'iza/utils.py'),
                           'iza.utils.list_exps': ('utils.html#list_exps', 'iza/utils.py'),
                           'iza.utils.load_exp_params': ('utils.html#load_exp_params', 'iza/utils.py'),
                           'iza.utils.make_missing_dirs': ('utils.html#make_missing_dirs', 'iza/utils.py'),
                           'iza.utils.make_temp_file': ('utils.html#make_temp_file', 'iza/utils.py'),
                           'iza.utils.path_to_str': ('utils.html#path_to_str', 'iza/utils.py'),
                           'iza.utils.rich_entry_fn': ('utils.html#rich_entry_fn', 'iza/utils.py'),
                           'iza.utils.rich_init_tree': ('utils.html#rich_init_tree', 'iza/utils.py'),
                           'iza.utils.save_exp_params': ('utils.html#save_exp_params', 'iza/utils.py'),
                           'iza.utils.setup_exp': ('utils.html#setup_exp', 'iza/utils.py'),
                           'iza.utils.sort_directory_first': ('utils.html#sort_directory_first', 'iza/utils.py'),
                           'iza.utils.sort_file_first': ('utils.html#sort_file_first', 'iza/utils.py'),
                           'iza.utils.str_just_alpha': ('utils.html#str_just_alpha', 'iza/utils.py'),
                           'iza.utils.str_just_numeric': ('utils.html#str_just_numeric', 'iza/utils.py'),
                           'iza.utils.str_part_end': ('utils.html#str_part_end', 'iza/utils.py'),
                           'iza.utils.str_part_mid': ('utils.html#str_part_mid', 'iza/utils.py'),
                           'iza.utils.str_part_pre': ('utils.html#str_part_pre', 'iza/utils.py'),
                           'iza.utils.str_split_get': ('utils.html#str_split_get', 'iza/utils.py'),
                           'iza.utils.str_to_path': ('utils.html#str_to_path', 'iza/utils.py'),
                           'iza.utils.stream_file': ('utils.html#stream_file', 'iza/utils.py'),
                           'iza.utils.strip_punc': ('utils.html#strip_punc', 'iza/utils.py'),
                           'iza.utils.to_abs_expanded': ('utils.html#to_abs_expanded', 'iza/utils.py'),
                           'iza.utils.to_ndarray': ('utils.html#to_ndarray', 'iza/utils.py'),
                           'iza.utils.undo_gz': ('utils.html#undo_gz', 'iza/utils.py'),
                           'iza.utils.undo_npmatrix': ('utils.html#undo_npmatrix', 'iza/utils.py'),
                           'iza.utils.undo_sparse': ('utils.html#undo_sparse', 'iza/utils.py'),
                           'iza.utils.urljoin': ('utils.html#urljoin', 'iza/utils.py'),
                           'iza.utils.walk_dir_tree': ('utils.html#walk_dir_tree', 'iza/utils.py'),
                           'iza.utils.wrangle_kwargs_for_class': ('utils.html#wrangle_kwargs_for_class', 'iza/utils.py'),
                           'iza.utils.wrangle_kwargs_for_func': ('utils.html#wrangle_kwargs_for_func', 'iza/utils.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_plots.ipynb.

# %% auto 0
__all__ = ['set_rc_params', 'scale_figsize_for_grid', 'setup_grid_figure', 'mpl_cur_cmap', 'mpl_cat_val',
           'convert_cmap_to_categorical', 'make_categorical_legend', 'hide_axes', 'triple_scatter']

# %% ../nbs/04_plots.ipynb 4
import os
import numpy as np, pandas as pd

import matplotlib as mpl, matplotlib.pyplot as plt, seaborn as sns
from matplotlib.colors import ListedColormap, LinearSegmentedColormap
from matplotlib.lines import Line2D

from typing import Optional, List, Union, Tuple

# %% ../nbs/04_plots.ipynb 5
from .types import ColorMap, ndarray
from .static import RC_PARAMS
from .utils import (filter_kwargs_for_func, filter_kwargs_for_class)

# %% ../nbs/04_plots.ipynb 6
def set_rc_params(params: Optional[dict] = None):
    param_dict = RC_PARAMS.copy()
    if params is not None:        
        param_dict.update(params)
    plt.rcParams.update(param_dict)

def scale_figsize_for_grid(
    figsize:Optional[Tuple[int, int]]=(12, 8), 
    n_cols:Optional[int]=1, n_rows:Optional[int]=1,
) -> Tuple[int, int]:
    grid_figsize = (figsize[0] * n_cols, figsize[1] * n_rows)
    return grid_figsize

def setup_grid_figure(
    ncols: Optional[int] = 1, nrows: Optional[int] = 1,
    figsize: Optional[Tuple[int, int]] = (12, 8), 
    dpi: Optional[int] = 300, hspace: Optional[float] = 0.3, wspace: Optional[float] = None,
    frameon: Optional[bool] = False, **kwargs
) -> Tuple[plt.Figure, plt.GridSpec]:
    
    grid_figsize = scale_figsize_for_grid(figsize, ncols, nrows)
    params = dict(
        ncols=ncols, nrows=nrows, figsize=grid_figsize,
        dpi=dpi, hspace=hspace, wspace=wspace, frameon=frameon
    )
    params.update(kwargs)
    
    f_params = filter_kwargs_for_func(plt.figure, **params)
    g_params = filter_kwargs_for_class(plt.GridSpec, **params)
    
    fig = plt.figure(None, **f_params)
    gspec = plt.GridSpec(**g_params)    
    return fig, gspec

def mpl_cur_cmap(cmap: Optional[ColorMap] = None) -> ColorMap:
    if cmap is None:
        fig = plt.gcf()
        cmap = fig.get_cmap() 
    return cmap

def mpl_cat_val(i: int, n: int) -> float:
    adj = 1 if n > 1 else 0
    return i / (n - adj)

def convert_cmap_to_categorical(values: List[str], cmap: Optional[ColorMap]) -> ListedColormap:
    unique = np.unique(values)
    n_vals = len(unique)
    
    # Generate a list of colors from the original colormap
    cmap = mpl_cur_cmap(cmap)
    colors = [cmap(mpl_cat_val(i, n_vals)) for i in range(n_vals)]
    
    # Create a new colormap from the list of colors
    new_cmap = ListedColormap(colors)
    
    return new_cmap

def make_categorical_legend(
    values: List[str], cmap: Optional[ColorMap], **kwargs
) -> mpl.legend.Legend:
    
    # NOTE: c will be the ground truth labels from pd.factorize (e.g. [0.0,0,0,1,0,1,1,1])    
    unique = sorted(np.unique(values))
    cmap = mpl_cur_cmap(cmap)
    cmap = convert_cmap_to_categorical(unique, cmap)

    line_opts = dict(marker='o', markersize=15)
    line_opts.update(kwargs)

    legend_elements = [        
        Line2D(
            [0], [0], color = cmap(i), markerfacecolor = cmap(i),
            label = label, **line_opts
        )
        for i, label in enumerate(unique)
    ]

    leg = plt.legend(handles=legend_elements)
    return leg

# %% ../nbs/04_plots.ipynb 7
def hide_axes(ax:mpl.axes.Axes) -> mpl.axes.Axes:
    ax.set_xlabel('')
    ax.set_ylabel('')
    ax.get_xaxis().get_major_formatter().set_scientific(False)
    ax.get_yaxis().get_major_formatter().set_scientific(False)
    kwargs = dict(bottom=False, left=False, labelbottom=False, labelleft=False)
    ax.tick_params(which='both', **kwargs)
    return ax

def triple_scatter(
    data: ndarray, ax: mpl.axes.Axes,
    c: Union[List[str], List[float]], s:List[float],
    cmap: Optional[ColorMap] = None,
    alpha: float = 1, marker: str = '.',
    point_size: float = 300, hide_axes: bool = True,
    frameon: bool = False,
):
    x, y = zip(*data)
    cmap = mpl_cur_cmap(cmap)
    
    # Handle size of points for bottom black layer, and middle white layer
    outline_width = (0.3, 0.05)
    bg_width, gap_width = outline_width
    
    point = np.sqrt(point_size)
    gap_size = (point + (point * gap_width) * 2) ** 2
    bg_size = (np.sqrt(gap_size) + (point * bg_width) * 2) ** 2

    scatter_params = dict(
        s = bg_size * np.sqrt(s),
        alpha = alpha, marker = marker, 
        linewidths = 0, edgecolors = None
    )

    scatter_params.update(c='black')
    ax.scatter(x, y, **scatter_params)

    scatter_params.update(c='white')
    ax.scatter(x, y, **scatter_params)

    scatter_params.update(
        c=c, s = point_size * np.sqrt(s),
        alpha = (0.7 * alpha),
        cmap = cmap,
    )    
    
    pnts = ax.scatter(x, y, **scatter_params)
    if hide_axes:
        ax = hide_axes(ax)

    ax.set_frame_on(frameon)
    ax.patch.set_alpha(0)
    return pnts
